import os
import json

from enum import Enum
from typing import List, Optional, Set

from gi.repository import NetworkManager, NMClient

from .from_json import from_json
from .config import AppConfiguration
from .cache_manager import CacheManager
from .server.api_objects import Child


class RepeatType(Enum):
    NO_REPEAT = 0
    REPEAT_QUEUE = 1
    REPEAT_SONG = 2

    @property
    def icon(self):
        icon_name = [
            'repeat',
            'repeat-symbolic',
            'repeat-song-symbolic',
        ][self.value]
        return 'media-playlist-' + icon_name

    def as_mpris_loop_status(self):
        return ['None', 'Playlist', 'Track'][self.value]

    @staticmethod
    def from_mpris_loop_status(loop_status):
        return {
            'None': RepeatType.NO_REPEAT,
            'Track': RepeatType.REPEAT_SONG,
            'Playlist': RepeatType.REPEAT_QUEUE,
        }[loop_status]


class ApplicationState:
    """
    Represents the state of the application. In general, there are two things
    that are stored here: configuration, and UI state.

    Configuration is stored in ``config`` which is an ``AppConfiguration``
    object. UI state is stored as separate properties on this class.

    Configuration is stored to disk in $XDG_CONFIG_HOME/sublime-music. State is
    stored in $XDG_CACHE_HOME. Nothing in state should be assumed to be
    permanent. State need not be saved, the ``to_json`` and ``from_json``
    functions define what part of the state will be saved across application
    loads.
    """
    version: int = 1
    config: AppConfiguration = AppConfiguration()
    config_file: Optional[str] = None
    playing: bool = False
    current_song_index: int = -1
    play_queue: List[str] = []
    old_play_queue: List[str] = []
    _volume: dict = {'this device': 100}
    is_muted: bool = False
    repeat_type: RepeatType = RepeatType.NO_REPEAT
    shuffle_on: bool = False
    song_progress: float = 0
    current_device: str = 'this device'
    current_tab: str = 'albums'
    selected_album_id: Optional[str] = None
    selected_artist_id: Optional[str] = None
    selected_browse_element_id: Optional[str] = None
    selected_playlist_id: Optional[str] = None

    # State for Album sort.
    current_album_sort: str = 'random'
    current_album_genre: str = 'Rock'
    current_album_alphabetical_sort: str = 'name'
    current_album_from_year: int = 2010
    current_album_to_year: int = 2020

    active_playlist_id: Optional[str] = None

    networkmanager_client = NMClient.Client.new()
    nmclient_initialized = False
    _current_ssids: Set[str] = set()

    def to_json(self):
        exclude = ('config', 'repeat_type', '_current_ssids')
        json_object = {
            k: getattr(self, k)
            for k in self.__annotations__.keys()
            if k not in exclude
        }
        json_object.update(
            {
                'repeat_type':
                getattr(self, 'repeat_type', RepeatType.NO_REPEAT).value,
            })
        return json_object

    def load_from_json(self, json_object):
        self.version = json_object.get('version', 0)
        self.current_song_index = json_object.get('current_song_index', -1)
        self.play_queue = json_object.get('play_queue', [])
        self.old_play_queue = json_object.get('old_play_queue', [])
        self._volume = json_object.get('_volume', {'this device': 100})
        self.is_muted = json_object.get('is_muted', False)
        self.repeat_type = RepeatType(json_object.get('repeat_type', 0))
        self.shuffle_on = json_object.get('shuffle_on', False)
        self.song_progress = json_object.get('song_progress', 0.0)
        self.current_device = json_object.get('current_device', 'this device')
        self.current_tab = json_object.get('current_tab', 'albums')
        self.selected_album_id = json_object.get('selected_album_id', None)
        self.selected_artist_id = json_object.get('selected_artist_id', None)
        self.selected_browse_element_id = json_object.get(
            'selected_browse_element_id', None)
        self.selected_playlist_id = json_object.get(
            'selected_playlist_id', None)
        self.current_album_sort = json_object.get(
            'current_album_sort', 'random')
        self.current_album_genre = json_object.get(
            'current_album_genre', 'Rock')
        self.current_album_alphabetical_sort = json_object.get(
            'current_album_alphabetical_sort', 'name')
        self.current_album_from_year = json_object.get(
            'current_album_from_year', 2010)
        self.current_album_to_year = json_object.get(
            'current_album_to_year', 2020)
        self.active_playlist_id = json_object.get('active_playlist_id', None)

    def load(self):
        self.config = self.get_config(self.config_file)

        if self.config.server is None:
            self.load_from_json({})
            self.migrate()
            return

        CacheManager.reset(self.config, self.config.server, self.current_ssids)

        if os.path.exists(self.state_filename):
            with open(self.state_filename, 'r') as f:
                try:
                    self.load_from_json(json.load(f))
                except json.decoder.JSONDecodeError:
                    # Who cares, it's just state.
                    self.load_from_json({})

        self.migrate()

    def migrate(self):
        """Use this function to migrate any state storage that has changed."""
        self.config.migrate()

    def save(self):
        # Make the necessary directories before writing the state.
        os.makedirs(os.path.dirname(self.state_filename), exist_ok=True)

        # Save the state
        with open(self.state_filename, 'w+') as f:
            f.write(json.dumps(self.to_json(), indent=2, sort_keys=True))

    def save_config(self):
        # Make the necessary directories before writing the config.
        os.makedirs(os.path.dirname(self.config_file), exist_ok=True)

        with open(self.config_file, 'w+') as f:
            f.write(
                json.dumps(self.config.to_json(), indent=2, sort_keys=True))

    def get_config(self, filename: str) -> AppConfiguration:
        if not os.path.exists(filename):
            return AppConfiguration()

        with open(filename, 'r') as f:
            try:
                return from_json(AppConfiguration, json.load(f))
            except json.decoder.JSONDecodeError:
                return AppConfiguration()

    @property
    def current_ssids(self):
        if not self.nmclient_initialized:
            # Only look at the active WiFi connections.
            for ac in self.networkmanager_client.get_active_connections():
                if ac.get_connection_type() != '802-11-wireless':
                    continue
                devs = ac.get_devices()
                if len(devs) != 1:
                    continue
                if devs[0].get_device_type() != NetworkManager.DeviceType.WIFI:
                    continue

                self._current_ssids.add(ac.get_id())

        return self._current_ssids

    @property
    def state_filename(self):
        default_cache_location = (
            os.environ.get('XDG_DATA_HOME')
            or os.path.expanduser('~/.local/share'))
        return os.path.join(
            default_cache_location,
            'sublime-music',
            CacheManager.calculate_server_hash(self.config.server),
            'state.yaml',
        )

    @property
    def current_song(self) -> Optional[Child]:
        if (not self.play_queue or self.current_song_index < 0
                or not CacheManager.ready()):
            return None

        current_song_id = self.play_queue[self.current_song_index]
        return CacheManager.get_song_details(current_song_id).result()

    @property
    def volume(self):
        return self._volume.get(self.current_device, 100)

    @volume.setter
    def volume(self, value):
        self._volume[self.current_device] = value



from typing import Union
from random import randint

import gi

gi.require_version('Gtk', '3.0')
from gi.repository import Gtk, GObject, Pango, GLib

from sublime.state_manager import ApplicationState
from sublime.cache_manager import CacheManager
from sublime.ui import util
from .icon_button import IconButton
from .spinner_image import SpinnerImage

from sublime.server.api_objects import (
    AlbumWithSongsID3,
    Child,
    Directory,
)


class AlbumWithSongs(Gtk.Box):
    __gsignals__ = {
        'song-selected': (
            GObject.SignalFlags.RUN_FIRST,
            GObject.TYPE_NONE,
            (),
        ),
        'song-clicked': (
            GObject.SignalFlags.RUN_FIRST,
            GObject.TYPE_NONE,
            (int, object, object),
        ),
    }

    def __init__(self, album, cover_art_size=200, show_artist_name=True):
        Gtk.Box.__init__(self, orientation=Gtk.Orientation.HORIZONTAL)
        self.album = album

        box = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)
        artist_artwork = SpinnerImage(
            loading=False,
            image_name='artist-album-list-artwork',
            spinner_name='artist-artwork-spinner',
            image_size=cover_art_size,
        )
        # Account for 10px margin on all sides with "+ 20".
        artist_artwork.set_size_request(
            cover_art_size + 20, cover_art_size + 20)
        box.pack_start(artist_artwork, False, False, 0)
        box.pack_start(Gtk.Box(), True, True, 0)
        self.pack_start(box, False, False, 0)

        def cover_art_future_done(f):
            artist_artwork.set_from_file(f.result())
            artist_artwork.set_loading(False)

        cover_art_filename_future = CacheManager.get_cover_art_filename(
            album.coverArt,
            before_download=lambda: artist_artwork.set_loading(True),
            size=cover_art_size,
        )
        cover_art_filename_future.add_done_callback(
            lambda f: GLib.idle_add(cover_art_future_done, f))

        album_details = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)
        album_title_and_buttons = Gtk.Box(
            orientation=Gtk.Orientation.HORIZONTAL)

        # TODO: deal with super long-ass titles
        album_title_and_buttons.add(
            Gtk.Label(
                label=album.get('name', album.get('title')),
                name='artist-album-list-album-name',
                halign=Gtk.Align.START,
                ellipsize=Pango.EllipsizeMode.END,
            ))

        self.play_btn = IconButton(
            'media-playback-start-symbolic', sensitive=False)
        self.play_btn.connect('clicked', self.play_btn_clicked)
        album_title_and_buttons.pack_start(self.play_btn, False, False, 5)

        self.shuffle_btn = IconButton(
            'media-playlist-shuffle-symbolic', sensitive=False)
        self.shuffle_btn.connect('clicked', self.shuffle_btn_clicked)
        album_title_and_buttons.pack_start(self.shuffle_btn, False, False, 5)

        self.play_next_btn = IconButton(
            'go-top-symbolic', action_name='app.play-next')
        album_title_and_buttons.pack_start(self.play_next_btn, False, False, 5)

        self.add_to_queue_btn = IconButton(
            'go-jump-symbolic', action_name='app.add-to-queue')
        album_title_and_buttons.pack_start(
            self.add_to_queue_btn, False, False, 5)

        self.download_all_btn = IconButton(
            'folder-download-symbolic', sensitive=False)
        self.download_all_btn.connect('clicked', self.on_download_all_click)
        album_title_and_buttons.pack_end(
            self.download_all_btn, False, False, 5)

        album_details.add(album_title_and_buttons)

        stats = [
            album.artist if show_artist_name else None,
            album.year,
            album.genre,

            # TODO when not available (not browse by tags), calculate after the
            # list loads and update the stats label.
            util.format_sequence_duration(album.duration)
            if album.get('duration') else None,
        ]

        album_details.add(
            Gtk.Label(
                label=util.dot_join(*stats),
                halign=Gtk.Align.START,
                margin_left=10,
            ))

        self.album_song_store = Gtk.ListStore(
            str,  # cache status
            str,  # title
            str,  # duration
            str,  # song ID
        )

        def create_column(header, text_idx, bold=False, align=0, width=None):
            renderer = Gtk.CellRendererText(
                xalign=align,
                weight=Pango.Weight.BOLD if bold else Pango.Weight.NORMAL,
                ellipsize=Pango.EllipsizeMode.END,
            )
            renderer.set_fixed_size(width or -1, 35)

            column = Gtk.TreeViewColumn(header, renderer, text=text_idx)
            column.set_resizable(True)
            column.set_expand(not width)
            return column

        self.loading_indicator = Gtk.Spinner(
            name='album-list-song-list-spinner')
        album_details.add(self.loading_indicator)

        self.album_songs = Gtk.TreeView(
            model=self.album_song_store,
            name='album-songs-list',
            headers_visible=False,
            margin_top=15,
            margin_left=10,
            margin_right=10,
            margin_bottom=10,
        )
        self.album_songs.get_selection().set_mode(Gtk.SelectionMode.MULTIPLE)

        # Song status column.
        renderer = Gtk.CellRendererPixbuf()
        renderer.set_fixed_size(30, 35)
        column = Gtk.TreeViewColumn('', renderer, icon_name=0)
        column.set_resizable(True)
        self.album_songs.append_column(column)

        self.album_songs.append_column(create_column('TITLE', 1, bold=True))
        self.album_songs.append_column(
            create_column('DURATION', 2, align=1, width=40))

        self.album_songs.connect('row-activated', self.on_song_activated)
        self.album_songs.connect(
            'button-press-event', self.on_song_button_press)
        self.album_songs.get_selection().connect(
            'changed', self.on_song_selection_change)
        album_details.add(self.album_songs)

        self.pack_end(album_details, True, True, 0)

        self.update_album_songs(album.id)

    # Event Handlers
    # =========================================================================
    def on_song_selection_change(self, event):
        if not self.album_songs.has_focus():
            self.emit('song-selected')

    def on_song_activated(self, treeview, idx, column):
        # The song ID is in the last column of the model.
        self.emit(
            'song-clicked',
            idx.get_indices()[0],
            [m[-1] for m in self.album_song_store],
            {},
        )

    def on_song_button_press(self, tree, event):
        if event.button == 3:  # Right click
            clicked_path = tree.get_path_at_pos(event.x, event.y)
            if not clicked_path:
                return False

            store, paths = tree.get_selection().get_selected_rows()
            allow_deselect = False

            def on_download_state_change(song_id=None):
                self.update_album_songs(self.album.id)

            # Use the new selection instead of the old one for calculating what
            # to do the right click on.
            if clicked_path[0] not in paths:
                paths = [clicked_path[0]]
                allow_deselect = True

            song_ids = [self.album_song_store[p][-1] for p in paths]

            # Used to adjust for the header row.
            bin_coords = tree.convert_tree_to_bin_window_coords(
                event.x, event.y)
            widget_coords = tree.convert_tree_to_widget_coords(
                event.x, event.y)

            util.show_song_popover(
                song_ids,
                event.x,
                event.y + abs(bin_coords.by - widget_coords.wy),
                tree,
                on_download_state_change=on_download_state_change,
            )

            # If the click was on a selected row, don't deselect anything.
            if not allow_deselect:
                return True

    def on_download_all_click(self, btn):
        CacheManager.batch_download_songs(
            [x[-1] for x in self.album_song_store],
            before_download=self.update,
            on_song_download_complete=lambda x: self.update(),
        )

    def play_btn_clicked(self, btn):
        song_ids = [x[-1] for x in self.album_song_store]
        self.emit(
            'song-clicked',
            0,
            song_ids,
            {'force_shuffle_state': False},
        )

    def shuffle_btn_clicked(self, btn):
        song_ids = [x[-1] for x in self.album_song_store]
        self.emit(
            'song-clicked',
            randint(0,
                    len(self.album_song_store) - 1),
            song_ids,
            {'force_shuffle_state': True},
        )

    # Helper Methods
    # =========================================================================
    def deselect_all(self):
        self.album_songs.get_selection().unselect_all()

    def update(self, force=False):
        self.update_album_songs(self.album.id)

    def set_loading(self, loading):
        if loading:
            self.loading_indicator.start()
            self.loading_indicator.show()
        else:
            self.loading_indicator.stop()
            self.loading_indicator.hide()

    @util.async_callback(
        lambda *a, **k: CacheManager.get_album(*a, **k),
        before_download=lambda self: self.set_loading(True),
        on_failure=lambda self, e: self.set_loading(False),
    )
    def update_album_songs(
            self,
            album: Union[AlbumWithSongsID3, Child, Directory],
            state: ApplicationState,
    ):
        new_store = [
            [
                util.get_cached_status_icon(
                    CacheManager.get_cached_status(song)),
                util.esc(song.title),
                util.format_song_duration(song.duration),
                song.id,
            ] for song in (album.get('child') or album.get('song') or [])
        ]

        song_ids = [song[-1] for song in new_store]

        self.play_btn.set_sensitive(True)
        self.shuffle_btn.set_sensitive(True)
        self.play_next_btn.set_action_target_value(
            GLib.Variant('as', song_ids))
        self.add_to_queue_btn.set_action_target_value(
            GLib.Variant('as', song_ids))
        self.download_all_btn.set_sensitive(True)

        util.diff_song_store(self.album_song_store, new_store)
        self.loading_indicator.hide()



import gi
gi.require_version('Gtk', '3.0')
from gi.repository import Gtk, GdkPixbuf


class SpinnerImage(Gtk.Overlay):
    def __init__(
            self,
            loading=True,
            image_name=None,
            spinner_name=None,
            image_size=None,
            **kwargs,
    ):
        Gtk.Overlay.__init__(self)
        self.image_size = image_size

        self.image = Gtk.Image(name=image_name, **kwargs)
        self.add(self.image)

        self.spinner = Gtk.Spinner(
            name=spinner_name,
            active=loading,
            halign=Gtk.Align.CENTER,
            valign=Gtk.Align.CENTER,
        )
        self.add_overlay(self.spinner)

    def set_from_file(self, filename):
        if filename == '':
            filename = None
        if self.image_size is not None and filename:
            pixbuf = GdkPixbuf.Pixbuf.new_from_file_at_scale(
                filename,
                self.image_size,
                self.image_size,
                True,
            )
            self.image.set_from_pixbuf(pixbuf)
        else:
            self.image.set_from_file(filename)

    def set_loading(self, loading_status):
        if loading_status:
            self.spinner.start()
            self.spinner.show()
        else:
            self.spinner.stop()
            self.spinner.hide()



#! /usr/bin/env python3
import sys

import gi
gi.require_version('Gtk', '3.0')
from gi.repository import Gtk  # noqa: F401

from .app import SublimeMusicApp


def main():
    app = SublimeMusicApp()
    app.run(sys.argv)



from datetime import datetime
from typing import Set

import gi
gi.require_version('Gtk', '3.0')
from gi.repository import Gio, Gtk, GObject, Gdk, GLib, Pango

from . import albums, artists, browse, playlists, player_controls
from sublime.state_manager import ApplicationState
from sublime.cache_manager import CacheManager, SearchResult
from sublime.server.api_objects import Child
from sublime.ui import util
from sublime.ui.common import SpinnerImage


class MainWindow(Gtk.ApplicationWindow):
    """Defines the main window for Sublime Music."""
    __gsignals__ = {
        'song-clicked': (
            GObject.SignalFlags.RUN_FIRST,
            GObject.TYPE_NONE,
            (int, object, object),
        ),
        'songs-removed': (
            GObject.SignalFlags.RUN_FIRST,
            GObject.TYPE_NONE,
            (object, ),
        ),
        'refresh-window': (
            GObject.SignalFlags.RUN_FIRST,
            GObject.TYPE_NONE,
            (object, bool),
        ),
        'go-to': (
            GObject.SignalFlags.RUN_FIRST,
            GObject.TYPE_NONE,
            (str, str),
        ),
    }

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.set_default_size(1150, 768)

        # Create the stack
        self.stack = self.create_stack(
            Albums=albums.AlbumsPanel(),
            Artists=artists.ArtistsPanel(),
            Browse=browse.BrowsePanel(),
            Playlists=playlists.PlaylistsPanel(),
        )
        self.stack.set_transition_type(
            Gtk.StackTransitionType.SLIDE_LEFT_RIGHT)

        self.titlebar = self.create_headerbar(self.stack)
        self.set_titlebar(self.titlebar)

        self.player_controls = player_controls.PlayerControls()
        self.player_controls.connect(
            'song-clicked', lambda _, *a: self.emit('song-clicked', *a))
        self.player_controls.connect(
            'songs-removed', lambda _, *a: self.emit('songs-removed', *a))
        self.player_controls.connect(
            'refresh-window',
            lambda _, *args: self.emit('refresh-window', *args),
        )

        flowbox = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)
        flowbox.pack_start(self.stack, True, True, 0)
        flowbox.pack_start(self.player_controls, False, True, 0)
        self.add(flowbox)

        self.connect('button-release-event', self.on_button_release)

    def update(self, state: ApplicationState, force=False):
        # Update the Connected to label on the popup menu.
        if state.config.current_server >= 0:
            server_name = state.config.servers[
                state.config.current_server].name
            self.connected_to_label.set_markup(
                f'<b>Connected to {server_name}</b>')
        else:
            self.connected_to_label.set_markup(
                f'<span style="italic">Not Connected to a Server</span>')

        self.stack.set_visible_child_name(state.current_tab)

        active_panel = self.stack.get_visible_child()
        if hasattr(active_panel, 'update'):
            active_panel.update(state, force=force)

        self.player_controls.update(state)

    def create_stack(self, **kwargs):
        stack = Gtk.Stack()
        for name, child in kwargs.items():
            child.connect(
                'song-clicked',
                lambda _, *args: self.emit('song-clicked', *args),
            )
            child.connect(
                'refresh-window',
                lambda _, *args: self.emit('refresh-window', *args),
            )
            stack.add_titled(child, name.lower(), name)
        return stack

    def create_headerbar(self, stack):
        """
        Configure the header bar for the window.
        """
        header = Gtk.HeaderBar()
        header.set_show_close_button(True)
        header.props.title = 'Sublime Music'

        # Search
        self.search_entry = Gtk.SearchEntry(
            placeholder_text='Search everything...')
        self.search_entry.connect('focus-in-event', self.on_search_entry_focus)
        self.search_entry.connect(
            'button-press-event', self.on_search_entry_button_press)
        self.search_entry.connect(
            'focus-out-event', self.on_search_entry_loose_focus)
        self.search_entry.connect('changed', self.on_search_entry_changed)
        self.search_entry.connect(
            'stop-search', self.on_search_entry_stop_search)
        header.pack_start(self.search_entry)

        # Search popup
        self.create_search_popup()

        # Stack switcher
        switcher = Gtk.StackSwitcher(stack=stack)
        header.set_custom_title(switcher)

        # Menu button
        menu_button = Gtk.MenuButton()
        menu_button.set_use_popover(True)
        menu_button.set_popover(self.create_menu())
        menu_button.connect('clicked', self.on_menu_clicked)
        self.menu.set_relative_to(menu_button)

        icon = Gio.ThemedIcon(name='open-menu-symbolic')
        image = Gtk.Image.new_from_gicon(icon, Gtk.IconSize.BUTTON)
        menu_button.add(image)

        header.pack_end(menu_button)

        return header

    def create_label(self, text, *args, **kwargs):
        label = Gtk.Label(
            use_markup=True,
            halign=Gtk.Align.START,
            ellipsize=Pango.EllipsizeMode.END,
            *args,
            **kwargs,
        )
        label.set_markup(text)
        label.get_style_context().add_class('search-result-row')
        return label

    def create_menu(self):
        self.menu = Gtk.PopoverMenu()

        self.connected_to_label = self.create_label(
            '', name='connected-to-label')
        self.connected_to_label.set_markup(
            f'<span style="italic">Not Connected to a Server</span>')

        menu_items = [
            (None, self.connected_to_label),
            (
                'app.configure-servers',
                Gtk.ModelButton(text='Configure Servers'),
            ),
            ('app.settings', Gtk.ModelButton(text='Settings')),
        ]

        vbox = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)
        for name, item in menu_items:
            if name:
                item.set_action_name(name)
            item.get_style_context().add_class('menu-button')
            vbox.pack_start(item, False, True, 0)
        self.menu.add(vbox)

        return self.menu

    def create_search_popup(self):
        self.search_popup = Gtk.PopoverMenu(modal=False)

        results_scrollbox = Gtk.ScrolledWindow(
            min_content_width=500,
            min_content_height=750,
        )

        def make_search_result_header(text):
            label = self.create_label(text)
            label.get_style_context().add_class('search-result-header')
            return label

        search_results_box = Gtk.Box(
            orientation=Gtk.Orientation.VERTICAL,
            name='search-results',
        )
        self.search_results_loading = Gtk.Spinner(
            active=False, name='search-spinner')
        search_results_box.add(self.search_results_loading)

        search_results_box.add(make_search_result_header('Songs'))
        self.song_results = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)
        search_results_box.add(self.song_results)

        search_results_box.add(make_search_result_header('Albums'))
        self.album_results = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)
        search_results_box.add(self.album_results)

        search_results_box.add(make_search_result_header('Artists'))
        self.artist_results = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)
        search_results_box.add(self.artist_results)

        search_results_box.add(make_search_result_header('Playlists'))
        self.playlist_results = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)
        search_results_box.add(self.playlist_results)

        results_scrollbox.add(search_results_box)
        self.search_popup.add(results_scrollbox)

        self.search_popup.set_relative_to(self.search_entry)
        rect = Gdk.Rectangle()
        rect.x = 22
        rect.y = 28
        rect.width = 1
        rect.height = 1
        self.search_popup.set_pointing_to(rect)
        self.search_popup.set_position(Gtk.PositionType.BOTTOM)

    # Event Listeners
    # =========================================================================
    def on_button_release(self, win, event):
        if not self.event_in_widgets(
                event,
                self.search_entry,
                self.search_popup,
        ):
            self.hide_search()

        if not self.event_in_widgets(
                event,
                self.player_controls.device_button,
                self.player_controls.device_popover,
        ):
            self.player_controls.device_popover.popdown()

        if not self.event_in_widgets(
                event,
                self.player_controls.play_queue_button,
                self.player_controls.play_queue_popover,
        ):
            self.player_controls.play_queue_popover.popdown()

        return False

    def on_menu_clicked(self, button):
        self.menu.popup()
        self.menu.show_all()

    def on_search_entry_focus(self, entry, event):
        self.show_search()

    def on_search_entry_button_press(self, *args):
        self.show_search()

    def on_search_entry_loose_focus(self, entry, event):
        self.hide_search()

    search_idx = 0
    latest_returned_search_idx = 0
    last_search_change_time = datetime.now()
    searches: Set[SearchResult] = set()

    def on_search_entry_changed(self, entry):
        now = datetime.now()
        if (now - self.last_search_change_time).seconds < 0.5:
            while len(self.searches) > 0:
                search = self.searches.pop()
                if search:
                    search.cancel()
        self.last_search_change_time = now

        if not self.search_popup.is_visible():
            self.search_popup.show_all()
            self.search_popup.popup()

        def create_search_callback(idx):
            def search_result_calback(result, is_last_in_batch):
                # Ignore slow returned searches.
                if idx < self.latest_returned_search_idx:
                    return

                # If all results are back, the stop the loading indicator.
                if is_last_in_batch:
                    if idx == self.search_idx - 1:
                        self.set_search_loading(False)
                    self.latest_returned_search_idx = idx

                self.update_search_results(result)

            return lambda *a: GLib.idle_add(search_result_calback, *a)

        self.searches.add(
            CacheManager.search(
                entry.get_text(),
                search_callback=create_search_callback(self.search_idx),
                before_download=lambda: self.set_search_loading(True),
            ))

        self.search_idx += 1

    def on_search_entry_stop_search(self, entry):
        self.search_popup.popdown()

    # Helper Functions
    # =========================================================================
    def show_search(self):
        self.search_entry.set_size_request(300, -1)
        self.search_popup.show_all()
        self.search_results_loading.hide()
        self.search_popup.popup()

    def hide_search(self):
        self.search_popup.popdown()
        self.search_entry.set_size_request(-1, -1)

    def set_search_loading(self, loading_state):
        if loading_state:
            self.search_results_loading.start()
            self.search_results_loading.show_all()
        else:
            self.search_results_loading.stop()
            self.search_results_loading.hide()

    def remove_all_from_widget(self, widget):
        for c in widget.get_children():
            widget.remove(c)

    def create_search_result_row(
        self,
        text,
        action_name,
        value,
        artwork_future,
    ):
        def on_search_row_button_press(btn, event):
            if action_name == 'song':
                goto_action_name, goto_id = 'album', value.albumId
            else:
                goto_action_name, goto_id = action_name, value.id
            self.emit('go-to', goto_action_name, goto_id)
            self.hide_search()

        row = Gtk.Button(relief=Gtk.ReliefStyle.NONE)
        row.connect('button-press-event', on_search_row_button_press)

        box = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL)
        image = SpinnerImage(image_name='search-artwork', image_size=30)
        box.add(image)
        box.add(self.create_label(text))
        row.add(box)

        def image_callback(f):
            image.set_loading(False)
            image.set_from_file(f.result())

        artwork_future.add_done_callback(
            lambda f: GLib.idle_add(image_callback, f))

        return row

    def update_search_results(self, search_results):
        # Songs
        if search_results.song is not None:
            self.remove_all_from_widget(self.song_results)
            for song in search_results.song or []:
                label_text = util.dot_join(
                    f'<b>{util.esc(song.title)}</b>',
                    util.esc(song.artist),
                )
                cover_art_future = CacheManager.get_cover_art_filename(
                    song.coverArt, size=50)
                self.song_results.add(
                    self.create_search_result_row(
                        label_text, 'song', song, cover_art_future))

            self.song_results.show_all()

        # Albums
        if search_results.album is not None:
            self.remove_all_from_widget(self.album_results)
            for album in search_results.album or []:
                name = album.title if type(album) == Child else album.name
                label_text = util.dot_join(
                    f'<b>{util.esc(name)}</b>',
                    util.esc(album.artist),
                )
                cover_art_future = CacheManager.get_cover_art_filename(
                    album.coverArt, size=50)
                self.album_results.add(
                    self.create_search_result_row(
                        label_text, 'album', album, cover_art_future))

            self.album_results.show_all()

        # Artists
        if search_results.artist is not None:
            self.remove_all_from_widget(self.artist_results)
            for artist in search_results.artist or []:
                label_text = util.esc(artist.name)
                cover_art_future = CacheManager.get_artist_artwork(artist)
                self.artist_results.add(
                    self.create_search_result_row(
                        label_text, 'artist', artist, cover_art_future))

            self.artist_results.show_all()

        # Playlists
        if search_results.playlist is not None:
            self.remove_all_from_widget(self.playlist_results)
            for playlist in search_results.playlist or []:
                label_text = util.esc(playlist.name)
                cover_art_future = CacheManager.get_cover_art_filename(
                    playlist.coverArt)
                self.playlist_results.add(
                    self.create_search_result_row(
                        label_text, 'playlist', playlist, cover_art_future))

            self.playlist_results.show_all()

    def event_in_widgets(self, event, *widgets):
        for widget in widgets:
            if not widget.is_visible():
                continue

            _, win_x, win_y = Gdk.Window.get_origin(self.get_window())
            widget_x, widget_y = widget.translate_coordinates(self, 0, 0)
            allocation = widget.get_allocation()

            bound_x = (win_x + widget_x, win_x + widget_x + allocation.width)
            bound_y = (win_y + widget_y, win_y + widget_y + allocation.height)

            # If the event is in this widget, return True immediately.
            if ((bound_x[0] <= event.x_root <= bound_x[1])
                    and (bound_y[0] <= event.y_root <= bound_y[1])):
                return True

        return False



from typing import Union

import gi

gi.require_version('Gtk', '3.0')
from gi.repository import Gtk, GObject, Pango, GLib, Gio

from sublime.state_manager import ApplicationState
from sublime.cache_manager import CacheManager
from sublime.ui import util
from sublime.ui.common import IconButton

from sublime.server.api_objects import Child, Artist


class BrowsePanel(Gtk.Overlay):
    """Defines the arist panel."""

    __gsignals__ = {
        'song-clicked': (
            GObject.SignalFlags.RUN_FIRST,
            GObject.TYPE_NONE,
            (int, object, object),
        ),
        'refresh-window': (
            GObject.SignalFlags.RUN_FIRST,
            GObject.TYPE_NONE,
            (object, bool),
        ),
    }

    id_stack = None

    def __init__(self):
        super().__init__()
        scrolled_window = Gtk.ScrolledWindow()

        self.root_directory_listing = ListAndDrilldown(IndexList)
        self.root_directory_listing.connect(
            'song-clicked',
            lambda _, *args: self.emit('song-clicked', *args),
        )
        self.root_directory_listing.connect(
            'refresh-window',
            lambda _, *args: self.emit('refresh-window', *args),
        )
        scrolled_window.add(self.root_directory_listing)

        self.add(scrolled_window)

        self.spinner = Gtk.Spinner(
            name='browse-spinner',
            active=True,
            halign=Gtk.Align.CENTER,
            valign=Gtk.Align.CENTER,
        )
        self.add_overlay(self.spinner)

    def update(self, state: ApplicationState, force=False):
        if not CacheManager.ready:
            return

        def do_update(id_stack):
            self.root_directory_listing.update(
                id_stack.result(),
                state=state,
                force=force,
            )
            self.spinner.hide()

        def calculate_path():
            if state.selected_browse_element_id is None:
                return []

            id_stack = []
            directory = None
            current_dir_id = state.selected_browse_element_id
            while directory is None or directory.parent is not None:
                directory = CacheManager.get_music_directory(
                    current_dir_id,
                    before_download=self.spinner.show,
                ).result()
                id_stack.append(directory.id)
                current_dir_id = directory.parent

            return id_stack

        path_fut = CacheManager.create_future(calculate_path)
        path_fut.add_done_callback(lambda f: GLib.idle_add(do_update, f))


class ListAndDrilldown(Gtk.Paned):
    __gsignals__ = {
        'song-clicked': (
            GObject.SignalFlags.RUN_FIRST,
            GObject.TYPE_NONE,
            (int, object, object),
        ),
        'refresh-window': (
            GObject.SignalFlags.RUN_FIRST,
            GObject.TYPE_NONE,
            (object, bool),
        ),
    }

    id_stack = None

    def __init__(self, list_type):
        Gtk.Paned.__init__(self, orientation=Gtk.Orientation.HORIZONTAL)

        self.list = list_type()
        self.list.connect(
            'song-clicked',
            lambda _, *args: self.emit('song-clicked', *args),
        )
        self.list.connect(
            'refresh-window',
            lambda _, *args: self.emit('refresh-window', *args),
        )
        self.pack1(self.list, False, False)

        self.drilldown = Gtk.Box()
        self.pack2(self.drilldown, True, False)

    def update(
        self,
        id_stack,
        state: ApplicationState,
        force=False,
        directory_id=None,
    ):
        self.list.update(
            None if len(id_stack) == 0 else id_stack[-1],
            state=state,
            force=force,
            directory_id=directory_id,
        )

        if self.id_stack == id_stack:
            # We always want to update, but in this case, we don't want to blow
            # away the drilldown.
            if isinstance(self.drilldown, ListAndDrilldown):
                self.drilldown.update(
                    id_stack[:-1],
                    state,
                    force=force,
                    directory_id=id_stack[-1],
                )
            return
        self.id_stack = id_stack

        if len(id_stack) > 0:
            self.remove(self.drilldown)
            self.drilldown = ListAndDrilldown(MusicDirectoryList)
            self.drilldown.connect(
                'song-clicked',
                lambda _, *args: self.emit('song-clicked', *args),
            )
            self.drilldown.connect(
                'refresh-window',
                lambda _, *args: self.emit('refresh-window', *args),
            )
            self.drilldown.update(
                id_stack[:-1],
                state,
                force=force,
                directory_id=id_stack[-1],
            )
            self.drilldown.show_all()
            self.pack2(self.drilldown, True, False)


class DrilldownList(Gtk.Box):
    __gsignals__ = {
        'song-clicked': (
            GObject.SignalFlags.RUN_FIRST,
            GObject.TYPE_NONE,
            (int, object, object),
        ),
        'refresh-window': (
            GObject.SignalFlags.RUN_FIRST,
            GObject.TYPE_NONE,
            (object, bool),
        ),
    }

    class DrilldownElement(GObject.GObject):
        id = GObject.Property(type=str)
        name = GObject.Property(type=str)
        is_dir = GObject.Property(type=bool, default=True)

        def __init__(self, element: Union[Child, Artist]):
            GObject.GObject.__init__(self)
            self.id = element.id
            self.name = (
                element.name if isinstance(element, Artist) else element.title)
            self.is_dir = element.get('isDir', True)

    def __init__(self):
        Gtk.Box.__init__(self, orientation=Gtk.Orientation.VERTICAL)

        list_actions = Gtk.ActionBar()

        refresh = IconButton('view-refresh-symbolic')
        refresh.connect('clicked', self.on_refresh_clicked)
        list_actions.pack_end(refresh)

        self.add(list_actions)

        self.loading_indicator = Gtk.ListBox()
        spinner_row = Gtk.ListBoxRow(activatable=False, selectable=False)
        spinner = Gtk.Spinner(name='drilldown-list-spinner', active=True)
        spinner_row.add(spinner)
        self.loading_indicator.add(spinner_row)
        self.pack_start(self.loading_indicator, False, False, 0)

        self.scroll_window = Gtk.ScrolledWindow(min_content_width=250)
        scrollbox = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)

        self.drilldown_directories_store = Gio.ListStore()
        self.list = Gtk.ListBox()
        self.list.bind_model(self.drilldown_directories_store, self.create_row)
        scrollbox.add(self.list)

        self.directory_song_store = Gtk.ListStore(
            str,  # cache status
            str,  # title
            str,  # duration
            str,  # song ID
        )

        def create_column(header, text_idx, bold=False, align=0, width=None):
            renderer = Gtk.CellRendererText(
                xalign=align,
                weight=Pango.Weight.BOLD if bold else Pango.Weight.NORMAL,
                ellipsize=Pango.EllipsizeMode.END,
            )
            renderer.set_fixed_size(width or -1, 35)

            column = Gtk.TreeViewColumn(header, renderer, text=text_idx)
            column.set_resizable(True)
            column.set_expand(not width)
            return column

        self.directory_song_list = Gtk.TreeView(
            model=self.directory_song_store,
            name='album-songs-list',
            headers_visible=False,
        )
        self.directory_song_list.get_selection().set_mode(
            Gtk.SelectionMode.MULTIPLE)

        # Song status column.
        renderer = Gtk.CellRendererPixbuf()
        renderer.set_fixed_size(30, 35)
        column = Gtk.TreeViewColumn('', renderer, icon_name=0)
        column.set_resizable(True)
        self.directory_song_list.append_column(column)

        self.directory_song_list.append_column(
            create_column('TITLE', 1, bold=True))
        self.directory_song_list.append_column(
            create_column('DURATION', 2, align=1, width=40))

        self.directory_song_list.connect(
            'row-activated', self.on_song_activated)
        self.directory_song_list.connect(
            'button-press-event', self.on_song_button_press)
        scrollbox.add(self.directory_song_list)

        self.scroll_window.add(scrollbox)
        self.pack_start(self.scroll_window, True, True, 0)

    def on_song_activated(self, treeview, idx, column):
        # The song ID is in the last column of the model.
        self.emit(
            'song-clicked',
            idx.get_indices()[0],
            [m[-1] for m in self.directory_song_store],
            {},
        )

    def on_song_button_press(self, tree, event):
        if event.button == 3:  # Right click
            clicked_path = tree.get_path_at_pos(event.x, event.y)
            if not clicked_path:
                return False

            store, paths = tree.get_selection().get_selected_rows()
            allow_deselect = False

            # Use the new selection instead of the old one for calculating what
            # to do the right click on.
            if clicked_path[0] not in paths:
                paths = [clicked_path[0]]
                allow_deselect = True

            song_ids = [self.directory_song_store[p][-1] for p in paths]

            # Used to adjust for the header row.
            bin_coords = tree.convert_tree_to_bin_window_coords(
                event.x, event.y)
            widget_coords = tree.convert_tree_to_widget_coords(
                event.x, event.y)

            util.show_song_popover(
                song_ids,
                event.x,
                event.y + abs(bin_coords.by - widget_coords.wy),
                tree,
                on_download_state_change=self.on_download_state_change,
            )

            # If the click was on a selected row, don't deselect anything.
            if not allow_deselect:
                return True

    def do_update_store(self, elements):
        new_directories_store = []
        new_songs_store = []
        selected_dir_idx = None

        for idx, el in enumerate(elements or []):
            if el.get('isDir', True):
                new_directories_store.append(
                    DrilldownList.DrilldownElement(el))
                if el.id == self.selected_id:
                    selected_dir_idx = idx
            else:
                new_songs_store.append(
                    [
                        util.get_cached_status_icon(
                            CacheManager.get_cached_status(el)),
                        util.esc(el.title),
                        util.format_song_duration(el.duration),
                        el.id,
                    ])

        util.diff_model_store(
            self.drilldown_directories_store, new_directories_store)

        util.diff_song_store(self.directory_song_store, new_songs_store)

        if len(new_directories_store) == 0:
            self.list.hide()
        else:
            self.list.show()

        if len(new_songs_store) == 0:
            self.directory_song_list.hide()
            self.scroll_window.set_min_content_width(275)
        else:
            self.directory_song_list.show()
            self.scroll_window.set_min_content_width(350)

        # Preserve selection
        if selected_dir_idx is not None:
            row = self.list.get_row_at_index(selected_dir_idx)
            self.list.select_row(row)

        self.loading_indicator.hide()

    def create_row(self, model: 'DrilldownList.DrilldownElement'):
        row = Gtk.ListBoxRow(
            action_name='app.browse-to',
            action_target=GLib.Variant('s', model.id),
        )
        rowbox = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL)
        rowbox.add(
            Gtk.Label(
                label=f'<b>{util.esc(model.name)}</b>',
                use_markup=True,
                margin=8,
                halign=Gtk.Align.START,
                ellipsize=Pango.EllipsizeMode.END,
            ))

        icon = Gio.ThemedIcon(name='go-next-symbolic')
        image = Gtk.Image.new_from_gicon(icon, Gtk.IconSize.BUTTON)
        rowbox.pack_end(image, False, False, 5)
        row.add(rowbox)
        row.show_all()
        return row


class IndexList(DrilldownList):
    def update(
        self,
        selected_id,
        state: ApplicationState = None,
        force=False,
        **kwargs,
    ):
        self.selected_id = selected_id
        self.update_store(force=force, state=state)

    def on_refresh_clicked(self, _):
        self.update(self.selected_id, force=True)

    @util.async_callback(
        lambda *a, **k: CacheManager.get_indexes(*a, **k),
        before_download=lambda self: self.loading_indicator.show(),
        on_failure=lambda self, e: self.loading_indicator.hide(),
    )
    def update_store(
        self,
        artists,
        state: ApplicationState = None,
        force=False,
    ):
        self.do_update_store(artists)

    def on_download_state_change(self, song_id=None):
        self.update(self.selected_id)


class MusicDirectoryList(DrilldownList):
    def update(
        self,
        selected_id,
        state: ApplicationState = None,
        force=False,
        directory_id=None,
    ):
        self.directory_id = directory_id
        self.selected_id = selected_id
        self.update_store(directory_id, force=force, state=state)

    def on_refresh_clicked(self, _):
        self.update(
            self.selected_id, force=True, directory_id=self.directory_id)

    @util.async_callback(
        lambda *a, **k: CacheManager.get_music_directory(*a, **k),
        before_download=lambda self: self.loading_indicator.show(),
        on_failure=lambda self, e: self.loading_indicator.hide(),
    )
    def update_store(
        self,
        directory,
        state: ApplicationState = None,
        force=False,
    ):
        self.do_update_store(directory.child)

    def on_download_state_change(self, song_id=None):
        self.update(self.selected_id, directory_id=self.directory_id)



import functools
import os
import re

from collections import defaultdict
from typing import Dict

from deepdiff import DeepDiff
from gi.repository import Gio, GLib

from .state_manager import RepeatType
from .cache_manager import CacheManager


def dbus_propagate(param_self=None):
    """
    Wraps a function which causes changes to DBus properties.
    """
    def decorator(function):
        @functools.wraps(function)
        def wrapper(*args):
            function(*args)
            (param_self or args[0]).dbus_manager.property_diff()

        return wrapper

    return decorator


class DBusManager:
    second_microsecond_conversion = 1000000

    current_state: Dict = {}

    def __init__(
        self,
        connection,
        do_on_method_call,
        on_set_property,
        get_state_and_player,
    ):
        self.get_state_and_player = get_state_and_player
        self.do_on_method_call = do_on_method_call
        self.on_set_property = on_set_property
        self.connection = connection

        def dbus_name_acquired(connection, name):
            specs = [
                'org.mpris.MediaPlayer2.xml',
                'org.mpris.MediaPlayer2.Player.xml',
                'org.mpris.MediaPlayer2.Playlists.xml',
                'org.mpris.MediaPlayer2.TrackList.xml',
            ]
            for spec in specs:
                spec_path = os.path.join(
                    os.path.dirname(__file__),
                    f'ui/mpris_specs/{spec}',
                )
                with open(spec_path) as f:
                    node_info = Gio.DBusNodeInfo.new_for_xml(f.read())

                connection.register_object(
                    '/org/mpris/MediaPlayer2',
                    node_info.interfaces[0],
                    self.on_method_call,
                    self.on_get_property,
                    self.on_set_property,
                )

        # TODO: I have no idea what to do here.
        def dbus_name_lost(*args):
            pass

        self.bus_number = Gio.bus_own_name_on_connection(
            connection,
            'org.mpris.MediaPlayer2.sublimemusic',
            Gio.BusNameOwnerFlags.NONE,
            dbus_name_acquired,
            dbus_name_lost,
        )

    def shutdown(self):
        Gio.bus_unown_name(self.bus_number)

    def on_get_property(
        self,
        connection,
        sender,
        path,
        interface,
        property_name,
    ):
        value = self.property_dict().get(interface, {}).get(property_name)
        return DBusManager.to_variant(value)

    def on_method_call(
        self,
        connection,
        sender,
        path,
        interface,
        method,
        params,
        invocation,
    ):
        # TODO I don't even know if this works.
        if interface == 'org.freedesktop.DBus.Properties':
            if method == 'Get':
                invocation.return_value(
                    self.on_get_property(
                        connection, sender, path, interface, *params))
            elif method == 'Set':
                self.on_set_property(
                    connection, sender, path, interface, *params)
            elif method == 'GetAll':
                all_properties = {
                    k: DBusManager.to_variant(v)
                    for k, v in self.property_dict()[interface].items()
                }
                invocation.return_value(
                    GLib.Variant('(a{sv})', (all_properties, )))

            return
        self.do_on_method_call(
            connection,
            sender,
            path,
            interface,
            method,
            params,
            invocation,
        )

    @staticmethod
    def to_variant(value):
        if callable(value):
            return DBusManager.to_variant(value())

        if isinstance(value, GLib.Variant):
            return value

        if type(value) == tuple:
            return GLib.Variant(*value)

        if type(value) == dict:
            return GLib.Variant(
                'a{sv}',
                {k: DBusManager.to_variant(v)
                 for k, v in value.items()},
            )

        variant_type = {
            list: 'as',
            str: 's',
            int: 'i',
            float: 'd',
            bool: 'b',
        }.get(type(value))
        if not variant_type:
            return value
        return GLib.Variant(variant_type, value)

    def property_dict(self):
        state, player = self.get_state_and_player()
        has_current_song = state.current_song is not None
        has_next_song = False
        if state.repeat_type in (RepeatType.REPEAT_QUEUE,
                                 RepeatType.REPEAT_SONG):
            has_next_song = True
        elif has_current_song:
            has_next_song = (
                state.current_song_index < len(state.play_queue) - 1)

        if state.active_playlist_id is None:
            active_playlist = (False, GLib.Variant('(oss)', ('/', '', '')))
        else:
            playlist_result = CacheManager.get_playlist(
                state.active_playlist_id)

            if playlist_result.is_future:
                # If we have to wait for the playlist result, just return
                # no playlist.
                active_playlist = (False, GLib.Variant('(oss)', ('/', '', '')))
            else:
                playlist = playlist_result.result()

                active_playlist = (
                    True,
                    GLib.Variant(
                        '(oss)',
                        (
                            '/playlist/' + playlist.id,
                            playlist.name,
                            CacheManager.get_cover_art_url(playlist.coverArt),
                        ),
                    ),
                )

        get_playlists_result = CacheManager.get_playlists()
        if get_playlists_result.is_future:
            playlist_count = 0
        else:
            playlist_count = len(get_playlists_result.result())

        return {
            'org.mpris.MediaPlayer2': {
                'CanQuit': True,
                'CanRaise': True,
                'HasTrackList': True,
                'Identity': 'Sublime Music',
                # TODO should implement in #29
                'DesktopEntry': 'foo',
                'SupportedUriSchemes': [],
                'SupportedMimeTypes': [],
            },
            'org.mpris.MediaPlayer2.Player': {
                'PlaybackStatus': {
                    (False, False): 'Stopped',
                    (False, True): 'Stopped',
                    (True, False): 'Paused',
                    (True, True): 'Playing',
                }[player is not None and player.song_loaded, state.playing],
                'LoopStatus':
                state.repeat_type.as_mpris_loop_status(),
                'Rate':
                1.0,
                'Shuffle':
                state.shuffle_on,
                'Metadata':
                self.get_mpris_metadata(
                    state.current_song_index,
                    state.play_queue,
                ) if state.current_song else {},
                'Volume':
                0.0 if state.is_muted else state.volume / 100,
                'Position': (
                    'x',
                    int(
                        max(state.song_progress or 0, 0)
                        * self.second_microsecond_conversion),
                ),
                'MinimumRate':
                1.0,
                'MaximumRate':
                1.0,
                'CanGoNext':
                has_current_song and has_next_song,
                'CanGoPrevious':
                has_current_song,
                'CanPlay':
                True,
                'CanPause':
                True,
                'CanSeek':
                True,
                'CanControl':
                True,
            },
            'org.mpris.MediaPlayer2.TrackList': {
                'Tracks': self.get_dbus_playlist(state.play_queue),
                'CanEditTracks': False,
            },
            'org.mpris.MediaPlayer2.Playlists': {
                # TODO this may do a network request. This really is a case for
                # doing the whole thing with caching some data beforehand.
                'PlaylistCount': playlist_count,
                'Orderings': ['Alphabetical', 'Created', 'Modified'],
                'ActivePlaylist': ('(b(oss))', active_playlist),
            },
        }

    def get_mpris_metadata(self, idx: int, play_queue):
        song_result = CacheManager.get_song_details(play_queue[idx])
        if song_result.is_future:
            return {}
        song = song_result.result()

        trackid = self.get_dbus_playlist(play_queue)[idx]
        duration = (
            'x',
            (song.duration or 0) * self.second_microsecond_conversion,
        )
        return {
            'mpris:trackid': trackid,
            'mpris:length': duration,
            'mpris:artUrl': CacheManager.get_cover_art_url(song.id, 1000),
            'xesam:album': song.album or '',
            'xesam:albumArtist': [song.artist or ''],
            'xesam:artist': [song.artist or ''],
            'xesam:title': song.title,
        }

    def get_dbus_playlist(self, play_queue):
        seen_counts = defaultdict(int)
        tracks = []
        for song_id in play_queue:
            id_ = seen_counts[song_id]
            tracks.append(f'/song/{song_id}/{id_}')
            seen_counts[song_id] += 1

        return tracks

    diff_parse_re = re.compile(r"root\['(.*?)'\]\['(.*?)'\](?:\[.*\])?")

    def property_diff(self):
        new_property_dict = self.property_dict()
        diff = DeepDiff(self.current_state, new_property_dict)

        changes = defaultdict(dict)

        for path, change in diff.get('values_changed', {}).items():
            interface, property_name = self.diff_parse_re.match(path).groups()
            changes[interface][property_name] = change['new_value']

        if diff.get('dictionary_item_added'):
            changes = new_property_dict

        for interface, changed_props in changes.items():
            # If the metadata has changed, just make the entire Metadata object
            # part of the update.
            if 'Metadata' in changed_props.keys():
                changed_props['Metadata'] = new_property_dict[interface][
                    'Metadata']

            # Special handling for when the position changes (a seek).
            # Technically, I'm sending this signal too often, but I don't think
            # it really matters.
            if (interface == 'org.mpris.MediaPlayer2.Player'
                    and 'Position' in changed_props):
                self.connection.emit_signal(
                    None,
                    '/org/mpris/MediaPlayer2',
                    interface,
                    'Seeked',
                    GLib.Variant('(x)', (changed_props['Position'][1], )),
                )

                # Do not emit the property change.
                del changed_props['Position']

            # Special handling for when the track list changes.
            # Technically, I'm supposed to use `TrackAdded` and `TrackRemoved`
            # signals when minor changes occur, but the docs also say that:
            #
            # > It is left up to the implementation to decide when a change to
            # > the track list is invasive enough that this signal should be
            # > emitted instead of a series of TrackAdded and TrackRemoved
            # > signals.
            #
            # So I think that any change is invasive enough that I should use
            # this signal.
            if (interface == 'org.mpris.MediaPlayer2.TrackList'
                    and 'Tracks' in changed_props):
                track_list = changed_props['Tracks']
                if len(track_list) > 0:
                    current_track = (
                        new_property_dict['org.mpris.MediaPlayer2.Player']
                        ['Metadata'].get('mpris:trackid', track_list[0]))
                    self.connection.emit_signal(
                        None,
                        '/org/mpris/MediaPlayer2',
                        interface,
                        'TrackListReplaced',
                        GLib.Variant('(aoo)', (track_list, current_track)),
                    )

            self.connection.emit_signal(
                None,
                '/org/mpris/MediaPlayer2',
                'org.freedesktop.DBus.Properties',
                'PropertiesChanged',
                GLib.Variant(
                    '(sa{sv}as)', (
                        interface,
                        {
                            k: DBusManager.to_variant(v)
                            for k, v in changed_props.items()
                        },
                        [],
                    )),
            )

        # Update state for next diff.
        self.current_state = new_property_dict



import math

from datetime import datetime
from pathlib import Path
from typing import List

import gi
gi.require_version('Gtk', '3.0')
from gi.repository import Gtk, GdkPixbuf, Pango, GObject, GLib

from sublime.cache_manager import CacheManager
from sublime.state_manager import ApplicationState, RepeatType
from sublime.ui import util
from sublime.ui.common import IconButton, SpinnerImage
from sublime.ui.common.players import ChromecastPlayer


class PlayerControls(Gtk.ActionBar):
    """
    Defines the player controls panel that appears at the bottom of the window.
    """
    __gsignals__ = {
        'song-scrub': (
            GObject.SignalFlags.RUN_FIRST,
            GObject.TYPE_NONE,
            (float, ),
        ),
        'volume-change': (
            GObject.SignalFlags.RUN_FIRST,
            GObject.TYPE_NONE,
            (float, ),
        ),
        'device-update': (
            GObject.SignalFlags.RUN_FIRST,
            GObject.TYPE_NONE,
            (str, ),
        ),
        'song-clicked': (
            GObject.SignalFlags.RUN_FIRST,
            GObject.TYPE_NONE,
            (int, object, object),
        ),
        'songs-removed': (
            GObject.SignalFlags.RUN_FIRST,
            GObject.TYPE_NONE,
            (object, ),
        ),
        'refresh-window': (
            GObject.SignalFlags.RUN_FIRST,
            GObject.TYPE_NONE,
            (object, bool),
        ),
    }
    editing: bool = False
    editing_play_queue_song_list: bool = False
    reordering_play_queue_song_list: bool = False
    current_song = None
    current_device = None
    chromecasts: List[ChromecastPlayer] = []

    def __init__(self):
        Gtk.ActionBar.__init__(self)
        self.set_name('player-controls-bar')

        song_display = self.create_song_display()
        playback_controls = self.create_playback_controls()
        play_queue_volume = self.create_play_queue_volume()

        self.last_device_list_update = None

        self.pack_start(song_display)
        self.set_center_widget(playback_controls)
        self.pack_end(play_queue_volume)

    def update(self, state: ApplicationState):
        self.current_device = state.current_device

        self.update_scrubber(
            getattr(state, 'song_progress', None),
            getattr(state.current_song, 'duration', None),
        )

        icon = 'pause' if state.playing else 'start'
        self.play_button.set_icon(f"media-playback-{icon}-symbolic")

        has_current_song = state.current_song is not None
        has_next_song = False
        if state.repeat_type in (RepeatType.REPEAT_QUEUE,
                                 RepeatType.REPEAT_SONG):
            has_next_song = True
        elif has_current_song:
            has_next_song = (
                state.current_song_index < len(state.play_queue) - 1)

        # Repeat button state
        # TODO: it's not correct to use symboloc vs. not symbolic icons for
        # lighter/darker versions of the icon. Fix this by using FG color I
        # think? But then we have to deal with styling, which sucks.
        self.repeat_button.set_icon(state.repeat_type.icon)

        # Shuffle button state
        # TODO: it's not correct to use symboloc vs. not symbolic icons for
        # lighter/darker versions of the icon. Fix this by using FG color I
        # think? But then we have to deal with styling, which sucks.
        self.shuffle_button.set_icon(
            'media-playlist-shuffle'
            + ('-symbolic' if state.shuffle_on else ''))

        self.song_scrubber.set_sensitive(has_current_song)
        self.prev_button.set_sensitive(has_current_song)
        self.play_button.set_sensitive(has_current_song)
        self.next_button.set_sensitive(has_current_song and has_next_song)

        # Volume button and slider
        if state.is_muted:
            icon_name = 'muted'
        elif state.volume < 30:
            icon_name = 'low'
        elif state.volume < 70:
            icon_name = 'medium'
        else:
            icon_name = 'high'

        self.volume_mute_toggle.set_icon(f'audio-volume-{icon_name}-symbolic')

        self.editing = True
        self.volume_slider.set_value(0 if state.is_muted else state.volume)
        self.editing = False

        # Update the current song information.
        # TODO add popup of bigger cover art photo here
        if state.current_song is not None:
            self.update_cover_art(state.current_song.coverArt, size='70')

            self.song_title.set_markup(util.esc(state.current_song.title))
            self.album_name.set_markup(util.esc(state.current_song.album))
            artist_name = util.esc(state.current_song.artist)
            self.artist_name.set_markup(artist_name or '')
        else:
            # Clear out the cover art and song tite if no song
            self.album_art.set_from_file(None)
            self.album_art.set_loading(False)
            self.song_title.set_markup('')
            self.album_name.set_markup('')
            self.artist_name.set_markup('')

        self.update_device_list()

        # Set the Play Queue button popup.
        if hasattr(state, 'play_queue'):
            play_queue_len = len(state.play_queue)
            if play_queue_len == 0:
                self.popover_label.set_markup('<b>Play Queue</b>')
            else:
                song_label = util.pluralize('song', play_queue_len)
                self.popover_label.set_markup(
                    f'<b>Play Queue:</b> {play_queue_len} {song_label}')

            self.editing_play_queue_song_list = True

            new_store = []

            def calculate_label(song_details):
                title = util.esc(song_details.title)
                album = util.esc(song_details.album)
                artist = util.esc(song_details.artist)
                return f'<b>{title}</b>\n{util.dot_join(album, artist)}'

            def make_idle_index_capturing_function(idx, fn):
                return lambda f: GLib.idle_add(fn, idx, f.result())

            def on_cover_art_future_done(idx, cover_art_filename):
                self.play_queue_store[idx][0] = cover_art_filename

            def on_song_details_future_done(idx, song_details):
                self.play_queue_store[idx][1] = calculate_label(song_details)

                # Cover Art
                cover_art_future = CacheManager.get_cover_art_filename(
                    song_details.coverArt,
                    size=50,
                )
                if cover_art_result.is_future:
                    # We don't have the cover art already cached.
                    cover_art_result.add_done_callback(
                        make_idle_index_capturing_function(
                            i,
                            on_cover_art_future_done,
                        ))
                else:
                    # We have the cover art already cached.
                    self.play_queue_store[idx][0] = cover_art_future.result()

            song_details_results = []
            for i, song_id in enumerate(state.play_queue):
                song_details_result = CacheManager.get_song_details(song_id)

                cover_art_filename = ''
                label = '\n'

                if song_details_result.is_future:
                    song_details_results.append((i, song_details_result))
                else:
                    # We have the details of the song already cached.
                    song_details = song_details_result.result()
                    label = calculate_label(song_details)

                    cover_art_result = CacheManager.get_cover_art_filename(
                        song_details.coverArt,
                        size=50,
                    )
                    if cover_art_result.is_future:
                        # We don't have the cover art already cached.
                        cover_art_result.add_done_callback(
                            make_idle_index_capturing_function(
                                i,
                                on_cover_art_future_done,
                            ))
                    else:
                        # We have the cover art already cached.
                        cover_art_filename = cover_art_result.result()

                new_store.append(
                    [
                        cover_art_filename,
                        label,
                        i == state.current_song_index,
                        song_id,
                    ])

            util.diff_song_store(self.play_queue_store, new_store)

            # Do this after the diff to avoid race conditions.
            for idx, song_details_result in song_details_results:
                song_details_result.add_done_callback(
                    make_idle_index_capturing_function(
                        idx,
                        on_song_details_future_done,
                    ))

            self.editing_play_queue_song_list = False

    @util.async_callback(
        lambda *k, **v: CacheManager.get_cover_art_filename(*k, **v),
        before_download=lambda self: self.album_art.set_loading(True),
        on_failure=lambda self, e: self.album_art.set_loading(False),
    )
    def update_cover_art(self, cover_art_filename, state):
        self.album_art.set_from_file(cover_art_filename)
        self.album_art.set_loading(False)

    def update_scrubber(self, current, duration):
        if current is None or duration is None:
            self.song_duration_label.set_text('-:--')
            self.song_progress_label.set_text('-:--')
            self.song_scrubber.set_value(0)
            return

        current = current or 0
        percent_complete = current / duration * 100

        if not self.editing:
            self.song_scrubber.set_value(percent_complete)
        self.song_duration_label.set_text(util.format_song_duration(duration))
        self.song_progress_label.set_text(
            util.format_song_duration(math.floor(current)))

    def on_scrub_state_change(self, scrubber_container, eventbutton):
        self.editing = not self.editing

        if not self.editing:
            self.emit('song-scrub', self.song_scrubber.get_value())

    def on_volume_change(self, scale):
        if not self.editing:
            self.emit('volume-change', scale.get_value())

    def on_play_queue_click(self, button):
        if self.play_queue_popover.is_visible():
            self.play_queue_popover.popdown()
        else:
            # TODO scroll the currently playing song into view.
            self.play_queue_popover.popup()
            self.play_queue_popover.show_all()

    def on_song_activated(self, treeview, idx, column):
        # The song ID is in the last column of the model.
        self.emit(
            'song-clicked',
            idx.get_indices()[0],
            [m[-1] for m in self.play_queue_store],
            {'no_reshuffle': True},
        )

    def update_device_list(self, force=False):
        self.device_list_loading.show()

        def chromecast_callback(chromecasts):
            self.chromecasts = chromecasts
            for c in self.chromecast_device_list.get_children():
                self.chromecast_device_list.remove(c)

            if self.current_device == 'this device':
                self.this_device.set_icon('audio-volume-high-symbolic')
            else:
                self.this_device.set_icon(None)

            chromecasts.sort(key=lambda c: c.device.friendly_name)
            for cc in chromecasts:
                icon = (
                    'audio-volume-high-symbolic'
                    if str(cc.device.uuid) == self.current_device else None)
                btn = IconButton(icon, label=cc.device.friendly_name)
                btn.get_style_context().add_class('menu-button')
                btn.connect(
                    'clicked',
                    lambda _, uuid: self.emit('device-update', uuid),
                    cc.device.uuid,
                )
                self.chromecast_device_list.add(btn)
                self.chromecast_device_list.show_all()

            self.device_list_loading.hide()
            self.last_device_list_update = datetime.now()

        update_diff = (
            None if not self.last_device_list_update else
            (datetime.now() - self.last_device_list_update).seconds)
        if (force or len(self.chromecasts) == 0
                or (self.last_device_list_update and update_diff > 60)):
            future = ChromecastPlayer.get_chromecasts()
            future.add_done_callback(
                lambda f: GLib.idle_add(chromecast_callback, f.result()))
        else:
            chromecast_callback(self.chromecasts)

    def on_device_click(self, button):
        if self.device_popover.is_visible():
            self.device_popover.popdown()
        else:
            self.device_popover.popup()
            self.device_popover.show_all()
            self.update_device_list()

    def on_device_refresh_click(self, button):
        self.update_device_list(force=True)

    def on_play_queue_button_press(self, tree, event):
        if event.button == 3:  # Right click
            clicked_path = tree.get_path_at_pos(event.x, event.y)

            store, paths = tree.get_selection().get_selected_rows()
            allow_deselect = False

            def on_download_state_change(song_id=None):
                # Refresh the entire window (no force) because the song could
                # be in a list anywhere in the window.
                self.emit('refresh-window', {}, False)

            # Use the new selection instead of the old one for calculating what
            # to do the right click on.
            if clicked_path[0] not in paths:
                paths = [clicked_path[0]]
                allow_deselect = True

            song_ids = [self.play_queue_store[p][-1] for p in paths]

            remove_text = (
                'Remove ' + util.pluralize('song', len(song_ids))
                + ' from queue')

            def on_remove_songs_click(_):
                self.emit('songs-removed', [p.get_indices()[0] for p in paths])

            util.show_song_popover(
                song_ids,
                event.x,
                event.y,
                tree,
                on_download_state_change=on_download_state_change,
                extra_menu_items=[
                    (Gtk.ModelButton(text=remove_text), on_remove_songs_click),
                ],
            )

            # If the click was on a selected row, don't deselect anything.
            if not allow_deselect:
                return True

    def on_play_queue_model_row_move(self, *args):
        # If we are programatically editing the song list, don't do anything.
        if self.editing_play_queue_song_list:
            return

        # We get both a delete and insert event, I think it's deterministic
        # which one comes first, but just in case, we have this
        # reordering_play_queue_song_list flag.
        if self.reordering_play_queue_song_list:
            currently_playing_index = [
                i for i, s in enumerate(self.play_queue_store) if s[2]
            ][0]
            self.emit(
                'refresh-window',
                {
                    'current_song_index': currently_playing_index,
                    'play_queue': [s[-1] for s in self.play_queue_store],
                },
                False,
            )
            self.reordering_play_queue_song_list = False
        else:
            self.reordering_play_queue_song_list = True

    def create_song_display(self):
        box = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL)

        self.album_art = SpinnerImage(
            image_name='player-controls-album-artwork')
        box.pack_start(self.album_art, False, False, 5)

        details_box = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)
        details_box.pack_start(Gtk.Box(), True, True, 0)

        def make_label(name):
            return Gtk.Label(
                name=name,
                halign=Gtk.Align.START,
                xalign=0,
                use_markup=True,
                ellipsize=Pango.EllipsizeMode.END,
            )

        self.song_title = make_label('song-title')
        details_box.add(self.song_title)

        self.album_name = make_label('album-name')
        details_box.add(self.album_name)

        self.artist_name = make_label('artist-name')
        details_box.add(self.artist_name)

        details_box.pack_start(Gtk.Box(), True, True, 0)
        box.pack_start(details_box, False, False, 5)

        return box

    def create_playback_controls(self):
        box = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)

        # Scrubber and song progress/length labels
        scrubber_box = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL)

        self.song_progress_label = Gtk.Label(label='-:--')
        scrubber_box.pack_start(self.song_progress_label, False, False, 5)

        self.song_scrubber = Gtk.Scale.new_with_range(
            orientation=Gtk.Orientation.HORIZONTAL, min=0, max=100, step=5)
        self.song_scrubber.set_name('song-scrubber')
        self.song_scrubber.set_draw_value(False)
        self.song_scrubber.connect(
            'button-press-event', self.on_scrub_state_change)
        self.song_scrubber.connect(
            'button-release-event', self.on_scrub_state_change)
        scrubber_box.pack_start(self.song_scrubber, True, True, 0)

        self.song_duration_label = Gtk.Label(label='-:--')
        scrubber_box.pack_start(self.song_duration_label, False, False, 5)

        box.add(scrubber_box)

        buttons = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL)
        buttons.pack_start(Gtk.Box(), True, True, 0)

        # Repeat button
        self.repeat_button = IconButton('media-playlist-repeat')
        self.repeat_button.set_action_name('app.repeat-press')
        buttons.pack_start(self.repeat_button, False, False, 5)

        # Previous button
        self.prev_button = IconButton(
            'media-skip-backward-symbolic',
            icon_size=Gtk.IconSize.LARGE_TOOLBAR)
        self.prev_button.set_action_name('app.prev-track')
        buttons.pack_start(self.prev_button, False, False, 5)

        # Play button
        self.play_button = IconButton(
            'media-playback-start-symbolic',
            relief=True,
            icon_size=Gtk.IconSize.LARGE_TOOLBAR)
        self.play_button.set_name('play-button')
        self.play_button.set_action_name('app.play-pause')
        buttons.pack_start(self.play_button, False, False, 0)

        # Next button
        self.next_button = IconButton(
            'media-skip-forward-symbolic',
            icon_size=Gtk.IconSize.LARGE_TOOLBAR)
        self.next_button.set_action_name('app.next-track')
        buttons.pack_start(self.next_button, False, False, 5)

        # Shuffle button
        self.shuffle_button = IconButton('media-playlist-shuffle')
        self.shuffle_button.set_action_name('app.shuffle-press')
        buttons.pack_start(self.shuffle_button, False, False, 5)

        buttons.pack_start(Gtk.Box(), True, True, 0)
        box.add(buttons)

        return box

    def create_play_queue_volume(self):
        vbox = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)
        vbox.pack_start(Gtk.Box(), True, True, 0)
        box = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL)

        # Device button (for chromecast)
        self.device_button = IconButton(
            'video-display-symbolic', icon_size=Gtk.IconSize.LARGE_TOOLBAR)
        self.device_button.connect('clicked', self.on_device_click)
        box.pack_start(self.device_button, False, True, 5)

        self.device_popover = Gtk.PopoverMenu(
            modal=False,
            name='device-popover',
        )
        self.device_popover.set_relative_to(self.device_button)

        device_popover_box = Gtk.Box(
            orientation=Gtk.Orientation.VERTICAL,
            name='device-popover-box',
        )
        device_popover_header = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL)

        self.popover_label = Gtk.Label(
            label='<b>Devices</b>',
            use_markup=True,
            halign=Gtk.Align.START,
            margin=5,
        )
        device_popover_header.add(self.popover_label)

        refresh_devices = IconButton('view-refresh-symbolic')
        refresh_devices.connect('clicked', self.on_device_refresh_click)
        device_popover_header.pack_end(refresh_devices, False, False, 0)

        device_popover_box.add(device_popover_header)

        device_list = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)

        self.this_device = IconButton(
            'audio-volume-high-symbolic',
            label='This Device',
        )
        self.this_device.get_style_context().add_class('menu-button')
        self.this_device.connect(
            'clicked', lambda *a: self.emit('device-update', 'this device'))
        device_list.add(self.this_device)

        device_list.add(Gtk.Separator(orientation=Gtk.Orientation.HORIZONTAL))

        self.device_list_loading = Gtk.Spinner(active=True)
        self.device_list_loading.get_style_context().add_class('menu-button')
        device_list.add(self.device_list_loading)

        self.chromecast_device_list = Gtk.Box(
            orientation=Gtk.Orientation.VERTICAL)
        device_list.add(self.chromecast_device_list)

        device_popover_box.pack_end(device_list, True, True, 0)

        self.device_popover.add(device_popover_box)

        # Play Queue button
        self.play_queue_button = IconButton(
            'view-list-symbolic', icon_size=Gtk.IconSize.LARGE_TOOLBAR)
        self.play_queue_button.connect('clicked', self.on_play_queue_click)
        box.pack_start(self.play_queue_button, False, True, 5)

        self.play_queue_popover = Gtk.PopoverMenu(
            modal=False,
            name='up-next-popover',
        )
        self.play_queue_popover.set_relative_to(self.play_queue_button)

        play_queue_popover_box = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)
        play_queue_popover_header = Gtk.Box(
            orientation=Gtk.Orientation.HORIZONTAL)

        self.popover_label = Gtk.Label(
            label='<b>Play Queue</b>',
            use_markup=True,
            halign=Gtk.Align.START,
            margin=10,
        )
        play_queue_popover_header.add(self.popover_label)

        load_play_queue = Gtk.Button(label='Load Queue from Server', margin=5)
        load_play_queue.set_action_name('app.update-play-queue-from-server')
        play_queue_popover_header.pack_end(load_play_queue, False, False, 0)

        play_queue_popover_box.add(play_queue_popover_header)

        play_queue_scrollbox = Gtk.ScrolledWindow(
            min_content_height=600,
            min_content_width=400,
        )

        self.play_queue_store = Gtk.ListStore(
            str,  # image filename
            str,  # title, album, artist
            bool,  # playing
            str,  # song ID
        )
        self.play_queue_list = Gtk.TreeView(
            model=self.play_queue_store,
            reorderable=True,
            headers_visible=False,
        )
        self.play_queue_list.get_selection().set_mode(
            Gtk.SelectionMode.MULTIPLE)

        # Album Art column.
        def filename_to_pixbuf(column, cell, model, iter, flags):
            filename = model.get_value(iter, 0)
            if not filename:
                cell.set_property('icon_name', '')
                return
            pixbuf = GdkPixbuf.Pixbuf.new_from_file(filename)

            # If this is the playing song, then overlay the play icon.
            if model.get_value(iter, 2):
                play_overlay_pixbuf = GdkPixbuf.Pixbuf.new_from_file(
                    str(
                        Path(__file__).parent.joinpath(
                            'images/play-queue-play.png')))

                play_overlay_pixbuf.composite(
                    pixbuf, 0, 0, 50, 50, 0, 0, 1, 1,
                    GdkPixbuf.InterpType.NEAREST, 255)

            cell.set_property('pixbuf', pixbuf)

        renderer = Gtk.CellRendererPixbuf()
        renderer.set_fixed_size(55, 60)
        column = Gtk.TreeViewColumn('', renderer)
        column.set_cell_data_func(renderer, filename_to_pixbuf)
        column.set_resizable(True)
        self.play_queue_list.append_column(column)

        renderer = Gtk.CellRendererText(
            markup=True,
            ellipsize=Pango.EllipsizeMode.END,
        )
        column = Gtk.TreeViewColumn('', renderer, markup=1)
        self.play_queue_list.append_column(column)

        self.play_queue_list.connect('row-activated', self.on_song_activated)
        self.play_queue_list.connect(
            'button-press-event', self.on_play_queue_button_press)

        # Set up drag-and-drop on the song list for editing the order of the
        # playlist.
        self.play_queue_store.connect(
            'row-inserted', self.on_play_queue_model_row_move)
        self.play_queue_store.connect(
            'row-deleted', self.on_play_queue_model_row_move)

        play_queue_scrollbox.add(self.play_queue_list)
        play_queue_popover_box.pack_end(play_queue_scrollbox, True, True, 0)

        self.play_queue_popover.add(play_queue_popover_box)

        # Volume mute toggle
        self.volume_mute_toggle = IconButton('audio-volume-high-symbolic')
        self.volume_mute_toggle.set_action_name('app.mute-toggle')
        box.pack_start(self.volume_mute_toggle, False, True, 0)

        # Volume slider
        self.volume_slider = Gtk.Scale.new_with_range(
            orientation=Gtk.Orientation.HORIZONTAL, min=0, max=100, step=5)
        self.volume_slider.set_name('volume-slider')
        self.volume_slider.set_draw_value(False)
        self.volume_slider.connect('value-changed', self.on_volume_change)
        box.pack_start(self.volume_slider, True, True, 0)

        vbox.pack_start(box, False, True, 0)
        vbox.pack_start(Gtk.Box(), True, True, 0)
        return vbox



import math
import os

from time import sleep
from urllib.parse import urlencode
from deprecated import deprecated
from typing import Optional, Dict, List, Union, cast
from datetime import datetime

import requests

from .api_objects import (
    AlbumInfo,
    AlbumList,
    AlbumList2,
    AlbumWithSongsID3,
    ArtistInfo,
    ArtistInfo2,
    ArtistsID3,
    ArtistWithAlbumsID3,
    Bookmarks,
    Child,
    Directory,
    Genres,
    Indexes,
    InternetRadioStations,
    License,
    Lyrics,
    MusicFolders,
    NowPlaying,
    Playlists,
    PlaylistWithSongs,
    PlayQueue,
    Response,
    ScanStatus,
    SearchResult,
    SearchResult2,
    SearchResult3,
    Shares,
    Starred,
    Starred2,
    Songs,
    User,
    Users,
    VideoInfo,
)


class Server:
    """
    Defines a *Sonic server.

    Notes:

    * The ``hls`` endpoint is not supported.
    * The ``getCaptions`` endpoint is not supported
    * None of the podcast endpoints are supported.
    * The ``jukeboxControl`` endpoint is not supported.
    * None of the chat message endpoints are supported.
    * The ``server`` module is stateless. The only thing that it does is allow
      the module's user to query the *sonic server via the API.
    """
    def __init__(
            self,
            name: str,
            hostname: str,
            username: str,
            password: str,
            disable_cert_verify: bool,
    ):
        self.name: str = name
        self.hostname: str = hostname
        self.username: str = username
        self.password: str = password
        self.disable_cert_verify: bool = disable_cert_verify

    def _get_params(self) -> Dict[str, str]:
        """See Subsonic API Introduction for details."""
        return dict(
            u=self.username,
            p=self.password,
            c='Sublime Music',
            f='json',
            v='1.15.0',
        )

    def _make_url(self, endpoint: str) -> str:
        return f'{self.hostname}/rest/{endpoint}.view'

    def _subsonic_error_to_exception(self, error) -> Exception:
        return Exception(f'{error.code}: {error.message}')

    # def _get(self, url, timeout=(3.05, 2), **params):
    def _get(self, url, **params):
        params = {**self._get_params(), **params}
        print(f'[START] get: {url}')

        if os.environ.get('SUBLIME_MUSIC_DEBUG_DELAY'):
            print(
                "SUBLIME_MUSIC_DEBUG_DELAY enabled. Pausing for",
                f"{os.environ['SUBLIME_MUSIC_DEBUG_DELAY']} seconds.",
            )
            sleep(int(os.environ['SUBLIME_MUSIC_DEBUG_DELAY']))

        # Deal with datetime parameters (convert to milliseconds since 1970)
        for k, v in params.items():
            if type(v) == datetime:
                params[k] = int(cast(datetime, v).timestamp() * 1000)

        result = requests.get(
            url,
            params=params,
            verify=not self.disable_cert_verify,
            # timeout=timeout,
        )
        # TODO make better
        if result.status_code != 200:
            raise Exception(f'[FAIL] get: {url} status={result.status_code}')

        print(f'[FINISH] get: {url}')
        return result

    def _get_json(
            self,
            url: str,
            **params: Union[None, str, datetime, int, List[int]],
    ) -> Response:
        """
        Make a get request to a *Sonic REST API. Handle all types of errors
        including *Sonic ``<error>`` responses.

        :returns: a Response containing all of the data of the
            response, deserialized
        :raises Exception: needs some work TODO
        """
        result = self._get(url, **params)
        subsonic_response = result.json()['subsonic-response']

        # TODO make better
        if not subsonic_response:
            raise Exception('Fail!')

        if subsonic_response['status'] == 'failed':
            code, message = (
                subsonic_response['error'].get('code'),
                subsonic_response['error'].get('message'),
            )
            raise Exception(f'Subsonic API Error #{code}: {message}')

        response = Response.from_json(subsonic_response)

        # Check for an error and if it exists, raise it.
        if response.get('error'):
            raise self._subsonic_error_to_exception(response.error)

        return response

    def do_download(self, url, **params) -> bytes:
        print('download', url)
        download = self._get(url, **params)
        if 'json' in download.headers.get('Content-Type'):
            # TODO make better
            raise Exception("Didn't expect JSON.")
        return download.content

    def ping(self) -> Response:
        """
        Used to test connectivity with the server.
        """
        return self._get_json(self._make_url('ping'))

    def get_license(self) -> License:
        """
        Get details about the software license.
        """
        result = self._get_json(self._make_url('getLicense'))
        return result.license

    def get_music_folders(self) -> MusicFolders:
        """
        Returns all configured top-level music folders.
        """
        result = self._get_json(self._make_url('getMusicFolders'))
        return result.musicFolders

    def get_indexes(
            self,
            music_folder_id: int = None,
            if_modified_since: int = None,
    ) -> Indexes:
        """
        Returns an indexed structure of all artists.

        :param music_folder_id: If specified, only return artists in the music
            folder with the given ID. See ``getMusicFolders``.
        :param if_modified_since: If specified, only return a result if the
            artist collection has changed since the given time.
        """
        result = self._get_json(
            self._make_url('getIndexes'),
            musicFolderId=music_folder_id,
            ifModifiedSince=if_modified_since)
        return result.indexes

    def get_music_directory(self, dir_id) -> Directory:
        """
        Returns a listing of all files in a music directory. Typically used
        to get list of albums for an artist, or list of songs for an album.

        :param dir_id: A string which uniquely identifies the music folder.
            Obtained by calls to ``getIndexes`` or ``getMusicDirectory``.
        """
        result = self._get_json(
            self._make_url('getMusicDirectory'), id=str(dir_id))
        return result.directory

    def get_genres(self) -> Genres:
        """
        Returns all genres.
        """
        result = self._get_json(self._make_url('getGenres'))
        return result.genres

    def get_artists(self, music_folder_id: int = None) -> ArtistsID3:
        """
        Similar to getIndexes, but organizes music according to ID3 tags.

        :param music_folder_id: If specified, only return artists in the music
            folder with the given ID. See ``getMusicFolders``.
        """
        result = self._get_json(
            self._make_url('getArtists'), musicFolderId=music_folder_id)
        return result.artists

    def get_artist(self, artist_id: int) -> ArtistWithAlbumsID3:
        """
        Returns details for an artist, including a list of albums. This method
        organizes music according to ID3 tags.

        :param artist_id: The artist ID.
        """
        result = self._get_json(self._make_url('getArtist'), id=artist_id)
        return result.artist

    def get_album(self, album_id: int) -> AlbumWithSongsID3:
        """
        Returns details for an album, including a list of songs. This method
        organizes music according to ID3 tags.

        :param album_id: The album ID.
        """
        result = self._get_json(self._make_url('getAlbum'), id=album_id)
        return result.album

    def get_song(self, song_id: int) -> Child:
        """
        Returns details for a song.

        :param song_id: The song ID.
        """
        result = self._get_json(self._make_url('getSong'), id=song_id)
        return result.song

    def get_videos(self) -> Optional[List[Child]]:
        """
        Returns all video files.
        """
        result = self._get_json(self._make_url('getVideos'))
        return result.videos.video

    def get_video_info(self, video_id: int) -> Optional[VideoInfo]:
        """
        Returns details for a video, including information about available
        audio tracks, subtitles (captions) and conversions.

        :param video_id: The video ID.
        """
        result = self._get_json(self._make_url('getVideoInfo'), id=video_id)
        return result.videoInfo

    def get_artist_info(
            self,
            id: int,
            count: int = None,
            include_not_present: bool = None,
    ) -> Optional[ArtistInfo]:
        """
        Returns artist info with biography, image URLs and similar artists,
        using data from last.fm.

        :param id: The artist, album, or song ID.
        :param count: Max number of similar artists to return. Defaults to 20,
            according to API Spec.
        :param include_not_present: Whether to return artists that are not
            present in the media library. Defaults to false according to API
            Spec.
        """
        result = self._get_json(
            self._make_url('getArtistInfo'),
            id=id,
            count=count,
            includeNotPresent=include_not_present,
        )
        return result.artistInfo

    def get_artist_info2(
            self,
            id: int,
            count: int = None,
            include_not_present: bool = None,
    ) -> Optional[ArtistInfo2]:
        """
        Similar to getArtistInfo, but organizes music according to ID3 tags.

        :param id: The artist, album, or song ID.
        :param count: Max number of similar artists to return. Defaults to 20,
            according to API Spec.
        :param include_not_present: Whether to return artists that are not
            present in the media library. Defaults to false according to API
            Spec.
        """
        result = self._get_json(
            self._make_url('getArtistInfo2'),
            id=id,
            count=count,
            includeNotPresent=include_not_present,
        )
        return result.artistInfo2

    def get_album_info(self, id: int) -> Optional[AlbumInfo]:
        """
        Returns album notes, image URLs etc, using data from last.fm.

        :param id: The album or song ID.
        """
        result = self._get_json(self._make_url('getAlbumInfo'), id=id)
        return result.albumInfo

    def get_album_info2(self, id: int) -> Optional[AlbumInfo]:
        """
        Similar to getAlbumInfo, but organizes music according to ID3 tags.

        :param id: The album or song ID.
        """
        result = self._get_json(self._make_url('getAlbumInfo2'), id=id)
        return result.albumInfo

    def get_similar_songs(self, id: int, count: int = None) -> List[Child]:
        """
        Returns a random collection of songs from the given artist and similar
        artists, using data from last.fm. Typically used for artist radio
        features.

        :param id: The artist, album or song ID.
        :param count: Max number of songs to return. Defaults to 50 according
            to API Spec.
        """
        result = self._get_json(
            self._make_url('getSimilarSongs'),
            id=id,
            count=count,
        )
        return result.similarSongs.song

    def get_similar_songs2(self, id: int, count: int = None) -> List[Child]:
        """
        Similar to getSimilarSongs, but organizes music according to ID3 tags.

        :param id: The artist, album or song ID.
        :param count: Max number of songs to return. Defaults to 50 according
            to API Spec.
        """
        result = self._get_json(
            self._make_url('getSimilarSongs2'),
            id=id,
            count=count,
        )
        return result.similarSongs2.song

    def get_top_songs(self, artist: str, count: int = None) -> List[Child]:
        """
        Returns top songs for the given artist, using data from last.fm.

        :param id: The artist name.
        :param count: Max number of songs to return. Defaults to 50 according
            to API Spec.
        """
        result = self._get_json(
            self._make_url('getTopSongs'),
            artist=artist,
            count=count,
        )
        return result.topSongs.song

    def get_album_list(
            self,
            type: str,
            size: int = None,
            offset: int = None,
            from_year: int = None,
            to_year: int = None,
            genre: str = None,
            music_folder_id: int = None,
    ) -> AlbumList:
        """
        Returns a list of random, newest, highest rated etc. albums. Similar to
        the album lists on the home page of the Subsonic web interface.

        :param type: The list type. Must be one of the following: ``random``,
            ``newest``, ``highest``, ``frequent``, ``recent``. Since 1.8.0 you
            can also use ``alphabeticalByName`` or ``alphabeticalByArtist`` to
            page through all albums alphabetically, and ``starred`` to retrieve
            starred albums.  Since 1.10.1 you can use ``byYear`` and
            ``byGenre`` to list albums in a given year range or genre.
        :param size: The number of albums to return. Max 500. Deafult is 10
            according to API Spec.
        :param offset: The list offset. Useful if you for example want to page
            through the list of newest albums. Default is 0 according to API
            Spec.
        :param from_year: Required if ``type`` is ``byYear``. The first year in
            the range. If ``fromYear > toYear`` a reverse chronological list is
            returned.
        :param to_year: Required if ``type`` is ``byYear``. The last year in
            the range.
        :param genre: Required if ``type`` is ``byGenre``. The name of the
            genre, e.g., "Rock".
        :param music_folder_id: (Since 1.11.0) Only return albums in the music
            folder with the given ID. See ``getMusicFolders``.
        """
        result = self._get_json(
            self._make_url('getAlbumList'),
            type=type,
            size=size,
            offset=offset,
            fromYear=from_year,
            toYear=to_year,
            genre=genre,
            musicFolderId=music_folder_id,
        )
        return result.albumList

    def get_album_list2(
            self,
            type: str,
            size: int = None,
            offset: int = None,
            from_year: int = None,
            to_year: int = None,
            genre: str = None,
            music_folder_id: int = None,
    ) -> AlbumList2:
        """
        Similar to getAlbumList, but organizes music according to ID3 tags.

        :param type: The list type. Must be one of the following: ``random``,
            ``newest``, ``frequent``, ``recent``, ``starred``,
            ``alphabeticalByName`` or ``alphabeticalByArtist``. Since 1.10.1
            you can use ``byYear`` and ``byGenre`` to list albums in a given
            year range or genre.
        :param size: The number of albums to return. Max 500. Deafult is 10
            according to API Spec.
        :param offset: The list offset. Useful if you for example want to page
            through the list of newest albums. Default is 0 according to API
            Spec.
        :param from_year: Required if ``type`` is ``byYear``. The first year in
            the range. If ``fromYear > toYear`` a reverse chronological list is
            returned.
        :param to_year: Required if ``type`` is ``byYear``. The last year in
            the range.
        :param genre: Required if ``type`` is ``byGenre``. The name of the
            genre, e.g., "Rock".
        :param music_folder_id: (Since 1.11.0) Only return albums in the music
            folder with the given ID. See ``getMusicFolders``.
        """
        result = self._get_json(
            self._make_url('getAlbumList2'),
            type=type,
            size=size,
            offset=offset,
            fromYear=from_year,
            toYear=to_year,
            genre=genre,
            musicFolderId=music_folder_id,
        )
        return result.albumList2

    def get_random_songs(
            self,
            size: int = None,
            genre: str = None,
            from_year: str = None,
            to_year: str = None,
            music_folder_id: int = None,
    ) -> Songs:
        """
        Returns random songs matching the given criteria.

        :param size: The maximum number of songs to return. Max 500. Defaults
            to 10 according to API Spec.
        :param genre: Only returns songs belonging to this genre.
        :param from_year: Only return songs published after or in this year.
        :param to_year: Only return songs published before or in this year.
        :param music_folder_id: Only return albums in the music folder with the
            given ID. See ``getMusicFolders``.
        """
        result = self._get_json(
            self._make_url('getRandomSongs'),
            size=size,
            genre=genre,
            fromYear=from_year,
            toYear=to_year,
            musicFolderId=music_folder_id,
        )
        return result.randomSongs

    def get_songs_by_genre(
            self,
            genre: str,
            count: int = None,
            offset: int = None,
            music_folder_id: int = None,
    ) -> Songs:
        """
        Returns songs in a given genre.

        :param genre: Only returns songs belonging to this genre.
        :param count: The maximum number of songs to return. Max 500. Defaults
            to 10 according to API Spec.
        :param offset: The offset. Useful if you want to page through the songs
            in a genre.
        :param music_folder_id: (Since 1.12.0) Only return albums in the music
            folder with the given ID. See ``getMusicFolders``.
        """
        result = self._get_json(
            self._make_url('getSongsByGenre'),
            genre=genre,
            count=count,
            offset=offset,
            musicFolderId=music_folder_id,
        )
        return result.songsByGenre

    def get_now_playing(self) -> NowPlaying:
        """
        Returns what is currently being played by all users. Takes no extra
        parameters.
        """
        result = self._get_json(self._make_url('getNowPlaying'))
        return result.nowPlaying

    def get_starred(self, music_folder_id: int = None) -> Starred:
        """
        Returns starred songs, albums and artists.

        :param music_folder_id: (Since 1.12.0) Only return results from the
            music folder with the given ID. See ``getMusicFolders``.
        """
        result = self._get_json(self._make_url('getStarred'))
        return result.starred

    def get_starred2(self, music_folder_id: int = None) -> Starred2:
        """
        Similar to getStarred, but organizes music according to ID3 tags.

        :param music_folder_id: (Since 1.12.0) Only return results from the
            music folder with the given ID. See ``getMusicFolders``.
        """
        result = self._get_json(self._make_url('getStarred2'))
        return result.starred2

    @deprecated(version='1.4.0', reason='You should use search2 instead.')
    def search(
            self,
            artist: str = None,
            album: str = None,
            title: str = None,
            any: str = None,
            count: int = None,
            offset: int = None,
            newer_than: datetime = None,
    ) -> SearchResult:
        """
        Returns a listing of files matching the given search criteria. Supports
        paging through the result.

        :param artist: Artist to search for.
        :param album: Album to searh for.
        :param title: Song title to search for.
        :param any: Searches all fields.
        :param count: Maximum number of results to return.
        :param offset: Search result offset. Used for paging.
        :param newer_than: Only return matches that are newer than this.
        """
        result = self._get_json(
            self._make_url('search'),
            artist=artist,
            album=album,
            title=title,
            any=any,
            count=count,
            offset=offset,
            newerThan=math.floor(newer_than.timestamp()
                                 * 1000) if newer_than else None,
        )
        return result.searchResult

    def search2(
            self,
            query: str,
            artist_count: int = None,
            artist_offset: int = None,
            album_count: int = None,
            album_offset: int = None,
            song_count: int = None,
            song_offset: int = None,
            music_folder_id: int = None,
    ) -> SearchResult2:
        """
        Returns albums, artists and songs matching the given search criteria.
        Supports paging through the result.

        :param query: Search query.
        :param artist_count: Maximum number of artists to return. Defaults to
            20 according to API Spec.
        :param artist_offset: Search result offset for artists. Used for
            paging. Defualts to 0 according to API Spec.
        :param album_count: Maximum number of albums to return. Defaults to 20
            according to API Spec.
        :param album_offset: Search result offset for albums. Used for paging.
            Defualts to 0 according to API Spec.
        :param song_count: Maximum number of songs to return. Defaults to 20
            according to API Spec.
        :param song_offset: Search result offset for songs. Used for paging.
            Defualts to 0 according to API Spec.
        :param music_folder_id: (Since 1.12.0) Only return results from the
            music folder with the given ID. See ``getMusicFolders``.
        """
        result = self._get_json(
            self._make_url('search2'),
            query=query,
            artistCount=artist_count,
            artistOffset=artist_offset,
            albumCount=album_count,
            albumOffset=album_offset,
            songCount=song_count,
            songOffset=song_offset,
            musicFolderId=music_folder_id,
        )
        return result.searchResult2

    def search3(
            self,
            query: str,
            artist_count: int = None,
            artist_offset: int = None,
            album_count: int = None,
            album_offset: int = None,
            song_count: int = None,
            song_offset: int = None,
            music_folder_id: int = None,
    ) -> SearchResult3:
        """
        Similar to search2, but organizes music according to ID3 tags.

        :param query: Search query.
        :param artist_count: Maximum number of artists to return. Defaults to
            20 according to API Spec.
        :param artist_offset: Search result offset for artists. Used for
            paging. Defualts to 0 according to API Spec.
        :param album_count: Maximum number of albums to return. Defaults to 20
            according to API Spec.
        :param album_offset: Search result offset for albums. Used for paging.
            Defualts to 0 according to API Spec.
        :param song_count: Maximum number of songs to return. Defaults to 20
            according to API Spec.
        :param song_offset: Search result offset for songs. Used for paging.
            Defualts to 0 according to API Spec.
        :param music_folder_id: (Since 1.12.0) Only return results from the
            music folder with the given ID. See ``getMusicFolders``.
        """
        result = self._get_json(
            self._make_url('search3'),
            query=query,
            artistCount=artist_count,
            artistOffset=artist_offset,
            albumCount=album_count,
            albumOffset=album_offset,
            songCount=song_count,
            songOffset=song_offset,
            musicFolderId=music_folder_id,
        )
        return result.searchResult3

    def get_playlists(self, username: str = None) -> Playlists:
        """
        Returns all playlists a user is allowed to play.

        :param username: (Since 1.8.0) If specified, return playlists for this
            user rather than for the authenticated user. The authenticated user
            must have admin role if this parameter is used.
        """
        result = self._get_json(
            self._make_url('getPlaylists'), username=username)
        return result.playlists

    def get_playlist(self, id: int = None) -> PlaylistWithSongs:
        """
        Returns a listing of files in a saved playlist.

        :param username: ID of the playlist to return, as obtained by
            ``getPlaylists``.
        """
        result = self._get_json(self._make_url('getPlaylist'), id=id)
        return result.playlist

    def create_playlist(
            self,
            playlist_id: int = None,
            name: str = None,
            song_id: Union[int, List[int]] = None,
    ) -> Union[PlaylistWithSongs, Response]:
        """
        Creates (or updates) a playlist.

        :param playlist_id: The playlist ID. Required if updating.
        :param name: The human-readable name of the playlist. Required if
            creating.
        :param song_id: ID(s) of a song in the playlist. Can be a single ID or
            a list of IDs.
        """
        result = self._get_json(
            self._make_url('createPlaylist'),
            playlistId=playlist_id,
            name=name,
            songId=song_id,
        )

        if result.playlist:
            return result.playlist
        else:
            return result

    def update_playlist(
            self,
            playlist_id: int,
            name: str = None,
            comment: str = None,
            public: bool = None,
            song_id_to_add: Union[int, List[int]] = None,
            song_index_to_remove: Union[int, List[int]] = None,
    ) -> Response:
        """
        Updates a playlist. Only the owner of a playlist is allowed to update
        it.

        :param playlist_id: The playlist ID. Required if updating.
        :param name: The human-readable name of the playlist.
        :param comment: The playlist comment.
        :param public: ``true`` if the playlist should be visible to all users,
            ``false`` otherwise.
        :param song_id_to_add: Add this song with this/these ID(s) to the
            playlist. Can be a single ID or a list of IDs.
        :param song_id_to_remove: Remove the song at this/these position(s) in
            the playlist. Can be a single ID or a list of IDs.
        """
        return self._get_json(
            self._make_url('updatePlaylist'),
            playlistId=playlist_id,
            name=name,
            comment=comment,
            public=public,
            songIdToAdd=song_id_to_add,
            songIndexToRemove=song_index_to_remove,
        )

    def delete_playlist(self, id: int) -> Response:
        """
        Deletes a saved playlist
        """
        return self._get_json(self._make_url('deletePlaylist'), id=id)

    def get_stream_url(
            self,
            id: str,
            max_bit_rate: int = None,
            format: str = None,
            time_offset: int = None,
            size: int = None,
            estimate_content_length: bool = False,
            converted: bool = False,
    ):
        """
        Gets the URL to streams a given file.

        :param id: A string which uniquely identifies the file to stream.
            Obtained by calls to ``getMusicDirectory``.
        :param maxBitRate: (Since 1.2.0) If specified, the server will attempt
            to limit the bitrate to this value, in kilobits per second. If set
            to zero, no limit is imposed.
        :param format: (Since 1.6.0) Specifies the preferred target format
            (e.g., "mp3" or "flv") in case there are multiple applicable
            transcodings. Starting with 1.9.0 you can use the special value
            "raw" to disable transcoding.
        :param timeOffset: Only applicable to video streaming. If specified,
            start streaming at the given offset (in seconds) into the video.
            Typically used to implement video skipping.
        :param size: (Since 1.6.0) Only applicable to video streaming.
            Requested video size specified as WxH, for instance "640x480".
        :param estimateContentLength: (Since 1.8.0). If set to ``True``, the
            *Content-Length* HTTP header will be set to an estimated value for
            transcoded or downsampled media. Defaults to False according to the
            API Spec.
        :param converted: (Since 1.14.0) Only applicable to video streaming.
            Subsonic can optimize videos for streaming by converting them to
            MP4. If a conversion exists for the video in question, then setting
            this parameter to ``True`` will cause the converted video to be
            returned instead of the original. Defaults to False according to
            the API Spec.
        """
        params = dict(
            **self._get_params(),
            id=id,
            maxBitRate=max_bit_rate,
            format=format,
            timeOffset=time_offset,
            size=size,
            estimateContentLength=estimate_content_length,
            converted=converted,
        )
        params = {k: v for k, v in params.items() if v}
        return self._make_url('stream') + '?' + urlencode(params)

    def download(self, id: str):
        """
        Downloads a given media file. Similar to stream, but this method
        returns the original media data without transcoding or downsampling.

        :param id: A string which uniquely identifies the file to stream.
            Obtained by calls to ``getMusicDirectory``.
        """
        return self.do_download(self._make_url('download'), id=id)

    def get_cover_art(self, id: str, size: str = None):
        """
        Returns the cover art image in binary form.

        :param id: The ID of a song, album or artist.
        :param size: If specified, scale image to this size.
        """
        return self.do_download(
            self._make_url('getCoverArt'), id=id, size=size)

    def get_cover_art_url(self, id: str, size: str = None):
        """
        Returns the cover art image in binary form.

        :param id: The ID of a song, album or artist.
        :param size: If specified, scale image to this size.
        """
        params = dict(**self._get_params(), id=id, size=size)
        params = {k: v for k, v in params.items() if v}
        return self._make_url('getCoverArt') + '?' + urlencode(params)

    def get_lyrics(self, artist: str = None, title: str = None) -> Lyrics:
        """
        Searches for and returns lyrics for a given song.

        :param artist: The artist name.
        :param title: The song title.
        """
        result = self._get_json(
            self._make_url('getLyrics'),
            artist=artist,
            title=title,
        )
        return result.lyrics

    def get_avatar(self, username: str):
        """
        Returns the avatar (personal image) for a user.

        :param username: the user in question.
        """
        return self.do_download(self._make_url('getAvatar'), username=username)

    def star(
            self,
            id: Union[int, List[int]] = None,
            album_id: Union[int, List[int]] = None,
            artist_id: Union[int, List[int]] = None,
    ) -> Response:
        """
        Attaches a star to a song, album or artist.

        :param id: The ID(s) of the file(s) (song(s)) or folder(s)
            (album(s)/artist(s)) to star. Can be a single ID or a list of IDs.
        :param album_id: The ID(s) of an album/albums to star. Use this rather
            than ``id`` if the client accesses the media collection according
            to ID3 tags rather than file structure. Can be a single ID or a
            list of IDs.
        :param artist_id: The ID(s) of an artist/artists to star. Use this
            rather than ``id`` if the client accesses the media collection
            according to ID3 tags rather than file structure. Can be a single
            ID or a list of IDs.
        """
        return self._get_json(
            self._make_url('star'),
            id=id,
            albumId=album_id,
            artistId=artist_id,
        )

    def unstar(
            self,
            id: Union[int, List[int]] = None,
            album_id: Union[int, List[int]] = None,
            artist_id: Union[int, List[int]] = None,
    ) -> Response:
        """
        Removes the star from a song, album or artist.

        :param id: The ID(s) of the file(s) (song(s)) or folder(s)
            (album(s)/artist(s)) to star. Can be a single ID or a list of IDs.
        :param album_id: The ID(s) of an album/albums to star. Use this rather
            than ``id`` if the client accesses the media collection according
            to ID3 tags rather than file structure. Can be a single ID or a
            list of IDs.
        :param artist_id: The ID(s) of an artist/artists to star. Use this
            rather than ``id`` if the client accesses the media collection
            according to ID3 tags rather than file structure. Can be a single
            ID or a list of IDs.
        """
        return self._get_json(
            self._make_url('unstar'),
            id=id,
            albumId=album_id,
            artistId=artist_id,
        )

    def set_rating(self, id: int, rating: int) -> Response:
        """
        Sets the rating for a music file.

        :param id: A string which uniquely identifies the file (song) or folder
            (album/artist) to rate.
        :param rating: The rating between 1 and 5 (inclusive), or 0 to remove
            the rating.
        """
        return self._get_json(
            self._make_url('setRating'), id=id, rating=rating)

    def scrobble(
            self,
            id: int,
            time: datetime = None,
            submission: bool = True,
    ) -> Response:
        """
        Registers the local playback of one or more media files. Typically used
        when playing media that is cached on the client. This operation
        includes the following:

            * "Scrobbles" the media files on last.fm if the user has configured
              his/her last.fm credentials on the Subsonic server (Settings >
              Personal).
            * Updates the play count and last played timestamp for the media
              files. (Since 1.11.0)
            * Makes the media files appear in the "Now playing" page in the web
              app, and appear in the list of songs returned by
              ``getNowPlaying`` (Since 1.11.0)

        Since 1.8.0 you may specify multiple id (and optionally time)
        parameters to scrobble multiple files.

        :param id: The ID of the file to scrobble.
        :param time: (Since 1.8.0) The time at which the song was listened to.
        :param submission: Whether this is a "submission" or a "now playing"
            notification.
        """
        return self._get_json(
            self._make_url('scrobble'),
            id=id,
            time=time,
            submission=submission,
        )

    def get_shares(self) -> Shares:
        """
        Returns information about shared media this user is allowed to manage.
        Takes no extra parameters.
        """
        result = self._get_json(self._make_url('getShares'))
        return result.shares

    def create_share(
            self,
            id: Union[int, List[int]],
            description: str = None,
            expires: datetime = None,
    ) -> Shares:
        """
        Creates a public URL that can be used by anyone to stream music or
        video from the Subsonic server. The URL is short and suitable for
        posting on Facebook, Twitter etc. Note: The user must be authorized to
        share (see Settings > Users > User is allowed to share files with
        anyone).

        :param id: ID(s) of (a) song(s), album(s) or video(s) to share. Can be
            a single ID or a list of IDs.
        :param description: A user-defined description that will be displayed
            to people visiting the shared media.
        :param expires: The time at which the share expires.
        """
        result = self._get_json(
            self._make_url('createShare'),
            id=id,
            description=description,
            expires=expires,
        )
        return result.shares

    def update_share(
            self,
            id: int,
            description: str = None,
            expires: datetime = None,
    ) -> Response:
        """
        Updates the description and/or expiration date for an existing share.

        :param id: ID of the share to update.
        :param description: A user-defined description that will be displayed
            to people visiting the shared media.
        :param expires: The time at which the share expires.
        """
        return self._get_json(
            self._make_url('updateShare'),
            id=id,
            description=description,
            expires=expires,
        )

    def delete_share(self, id: int) -> Response:
        """
        Deletes an existing share.

        :param id: ID of the share to delete.
        """
        return self._get_json(self._make_url('deleteShare'), id=id)

    def get_internet_radio_stations(self) -> InternetRadioStations:
        """
        Returns all internet radio stations. Takes no extra parameters.
        """
        result = self._get_json(self._make_url('getInternetRadioStations'))
        return result.internetRadioStations

    def create_internet_radio_station(
            self,
            stream_url: str,
            name: str,
            homepage_url: str = None,
    ) -> Response:
        """
        Adds a new internet radio station. Only users with admin privileges are
        allowed to call this method.

        :param stream_url: The stream URL for the station.
        :param name: The user-defined name for the station.
        :param homepage_url: The home page URL for the station.
        """
        return self._get_json(
            self._make_url('createInternetRadioStation'),
            streamUrl=stream_url,
            name=name,
            homepageUrl=homepage_url,
        )

    def update_internet_radio_station(
            self,
            id: int,
            stream_url: str,
            name: str,
            homepage_url: str = None,
    ) -> Response:
        """
        Updates an existing internet radio station. Only users with admin
        privileges are allowed to call this method.

        :param id: The ID for the station.
        :param stream_url: The stream URL for the station.
        :param name: The user-defined name for the station.
        :param homepage_url: The home page URL for the station.
        """
        return self._get_json(
            self._make_url('updateInternetRadioStation'),
            id=id,
            streamUrl=stream_url,
            name=name,
            homepageUrl=homepage_url,
        )

    def delete_internet_radio_station(self, id: int) -> Response:
        """
        Deletes an existing internet radio station. Only users with admin
        privileges are allowed to call this method.

        :param id: The ID for the station.
        """
        return self._get_json(
            self._make_url('deleteInternetRadioStation'), id=id)

    def get_user(self, username: str) -> User:
        """
        Get details about a given user, including which authorization roles and
        folder access it has. Can be used to enable/disable certain features in
        the client, such as jukebox control.

        :param username: The name of the user to retrieve. You can only
            retrieve your own user unless you have admin privileges.
        """
        result = self._get_json(self._make_url('getUser'), username=username)
        return result.user

    def get_users(self) -> Users:
        """
        Get details about all users, including which authorization roles and
        folder access they have. Only users with admin privileges are allowed
        to call this method.
        """
        result = self._get_json(self._make_url('getUsers'))
        return result.users

    def create_user(
            self,
            username: str,
            password: str,
            email: str,
            ldap_authenticated: bool = False,
            admin_role: bool = False,
            settings_role: bool = True,
            stream_role: bool = True,
            jukebox_role: bool = False,
            download_role: bool = False,
            upload_role: bool = False,
            playlist_role: bool = False,
            covert_art_role: bool = False,
            comment_role: bool = False,
            podcast_role: bool = False,
            share_role: bool = False,
            video_conversion_role: bool = False,
            music_folder_id: Union[int, List[int]] = None,
    ) -> Response:
        """
        Creates a new Subsonic user.

        :param username: The name of the new user.
        :param password: The password of the new user, either in clear text or
            hex-encoded.
        :param email: The email address of the new user.
        :param ldap_authenticated: Whether the user is authenicated in LDAP.
        :param admin_role: Whether the user is administrator.
        :param settings_role: Whether the user is allowed to change personal
            settings and password.
        :param stream_role: Whether the user is allowed to play files.
        :param jukebox_role: Whether the user is allowed to play files in
            jukebox mode.
        :param download_role: Whether the user is allowed to download files.
        :param upload_role: Whether the user is allowed to upload files.
        :param playlist_role: Whether the user is allowed to create and delete
            playlists. Since 1.8.0, changing this role has no effect.
        :param covert_art_role: Whether the user is allowed to change cover art
            and tags.
        :param comment_role: Whether the user is allowed to create and edit
            comments and ratings.
        :param podcast_role: Whether the user is allowed to administrate
            Podcasts.
        :param share_role: (Since 1.8.0) Whether the user is allowed to share
            files with anyone.
        :param video_conversion_role: (Since 1.15.0) Whether the user is
            allowed to start video conversions.
        :param music_folder_id: (Since 1.12.0) IDs of the music folders the
            user is allowed access to. Can be a single ID or a list of IDs.
        """
        return self._get_json(
            self._make_url('createUser'),
            username=username,
            password=password,
            email=email,
            ldapAuthenticated=ldap_authenticated,
            adminRole=admin_role,
            settingsRole=settings_role,
            streamRole=stream_role,
            jukeboxRole=jukebox_role,
            downloadRole=download_role,
            uploadRole=upload_role,
            playlistRole=playlist_role,
            coverArtRole=covert_art_role,
            commentRole=comment_role,
            podcastRole=podcast_role,
            shareRole=share_role,
            videoConversionRole=video_conversion_role,
            musicFolderId=music_folder_id,
        )

    def update_user(
            self,
            username: str,
            password: str = None,
            email: str = None,
            ldap_authenticated: bool = False,
            admin_role: bool = False,
            settings_role: bool = True,
            stream_role: bool = True,
            jukebox_role: bool = False,
            download_role: bool = False,
            upload_role: bool = False,
            playlist_role: bool = False,
            covert_art_role: bool = False,
            comment_role: bool = False,
            podcast_role: bool = False,
            share_role: bool = False,
            video_conversion_role: bool = False,
            music_folder_id: Union[int, List[int]] = None,
    ) -> Response:
        """
        Modifies an existing Subsonic user.

        :param username: The name of the user.
        :param password: The password of the user, either in clear text or
            hex-encoded.
        :param email: The email address of the user.
        :param ldap_authenticated: Whether the user is authenicated in LDAP.
        :param admin_role: Whether the user is administrator.
        :param settings_role: Whether the user is allowed to change personal
            settings and password.
        :param stream_role: Whether the user is allowed to play files.
        :param jukebox_role: Whether the user is allowed to play files in
            jukebox mode.
        :param download_role: Whether the user is allowed to download files.
        :param upload_role: Whether the user is allowed to upload files.
        :param playlist_role: Whether the user is allowed to create and delete
            playlists. Since 1.8.0, changing this role has no effect.
        :param covert_art_role: Whether the user is allowed to change cover art
            and tags.
        :param comment_role: Whether the user is allowed to create and edit
            comments and ratings.
        :param podcast_role: Whether the user is allowed to administrate
            Podcasts.
        :param share_role: (Since 1.8.0) Whether the user is allowed to share
            files with anyone.
        :param video_conversion_role: (Since 1.15.0) Whether the user is
            allowed to start video conversions.
        :param music_folder_id: (Since 1.12.0) IDs of the music folders the
            user is allowed access to. Can be a single ID or a list of IDs.
        """
        return self._get_json(
            self._make_url('updateUser'),
            username=username,
            password=password,
            email=email,
            ldapAuthenticated=ldap_authenticated,
            adminRole=admin_role,
            settingsRole=settings_role,
            streamRole=stream_role,
            jukeboxRole=jukebox_role,
            downloadRole=download_role,
            uploadRole=upload_role,
            playlistRole=playlist_role,
            coverArtRole=covert_art_role,
            commentRole=comment_role,
            podcastRole=podcast_role,
            shareRole=share_role,
            videoConversionRole=video_conversion_role,
            musicFolderId=music_folder_id,
        )

    def delete_user(self, username: str) -> Response:
        """
        Deletes an existing Subsonic user.

        :param username: The name of the new user.
        """
        return self._get_json(self._make_url('deleteUser'), username=username)

    def change_password(self, username: str, password: str) -> Response:
        """
        Changes the password of an existing Subsonic user. You can only change
        your own password unless you have admin privileges.

        :param username: The name of the user which should change its password.
        :param password: The new password of the new user, either in clear text
            of hex-encoded.
        """
        return self._get_json(
            self._make_url('changePassword'),
            username=username,
            password=password,
        )

    def get_bookmarks(self) -> Bookmarks:
        """
        Returns all bookmarks for this user. A bookmark is a position within a
        certain media file.
        """
        result = self._get_json(self._make_url('getBookmarks'))
        return result.bookmarks

    def create_bookmarks(
            self,
            id: int,
            position: int,
            comment: str = None,
    ) -> Response:
        """
        Creates or updates a bookmark (a position within a media file).
        Bookmarks are personal and not visible to other users.

        :param id: ID of the media file to bookmark. If a bookmark already
            exists for this file it will be overwritten.
        :param position: The position (in milliseconds) within the media file.
        :param comment: A user-defined comment.
        """
        return self._get_json(
            self._make_url('createBookmark'),
            id=id,
            position=position,
            comment=comment,
        )

    def delete_bookmark(self, id: int) -> Response:
        """
        Deletes the bookmark for a given file.

        :param id: ID of the media file for which to delete the bookmark. Other
            users' bookmarks are not affected.
        """
        return self._get_json(self._make_url('deleteBookmark'), id=id)

    def get_play_queue(self) -> Optional[PlayQueue]:
        """
        Returns the state of the play queue for this user (as set by
        ``savePlayQueue``). This includes the tracks in the play queue, the
        currently playing track, and the position within this track. Typically
        used to allow a user to move between different clients/apps while
        retaining the same play queue (for instance when listening to an audio
        book).
        """
        result = self._get_json(self._make_url('getPlayQueue'))
        return result.playQueue

    def save_play_queue(
            self,
            id: Union[int, List[int]],
            current: int = None,
            position: int = None,
    ) -> Response:
        """
        Saves the state of the play queue for this user. This includes the
        tracks in the play queue, the currently playing track, and the position
        within this track. Typically used to allow a user to move between
        different clients/apps while retaining the same play queue (for
        instance when listening to an audio book).

        :param id: ID(s) of a/the song(s) in the play queue. Can be either a
            single ID or a list of IDs.
        :param current: The ID of the current playing song.
        :param position: The position in milliseconds within the currently
            playing song.
        """
        return self._get_json(
            self._make_url('savePlayQueue'),
            id=id,
            current=current,
            position=position,
        )

    def get_scan_status(self) -> ScanStatus:
        """
        Returns the current status for media library scanning. Takes no extra
        parameters.
        """
        result = self._get_json(self._make_url('getScanStatus'))
        return result.scanStatus

    def start_scan(self) -> ScanStatus:
        """
        Initiates a rescan of the media libraries. Takes no extra parameters.
        """
        result = self._get_json(self._make_url('startScan'))
        return result.scanStatus



import functools
from typing import Callable, List, Tuple, Any
import re

from concurrent.futures import Future

from deepdiff import DeepDiff

import gi
gi.require_version('Gtk', '3.0')
from gi.repository import Gtk, GLib, Gdk

from sublime.cache_manager import CacheManager, SongCacheStatus


def format_song_duration(duration_secs) -> str:
    """
    Formats the song duration as mins:seconds with the seconds being
    zero-padded if necessary.

    >>> format_song_duration(80)
    '1:20'
    >>> format_song_duration(62)
    '1:02'
    """
    return f'{duration_secs // 60}:{duration_secs % 60:02}'


def pluralize(string: str, number: int, pluralized_form=None):
    """
    Pluralize the given string given the count as a number.

    >>> pluralize('foo', 1)
    'foo'
    >>> pluralize('foo', 2)
    'foos'
    >>> pluralize('foo', 0)
    'foos'
    """
    if number != 1:
        return pluralized_form or f'{string}s'
    return string


def format_sequence_duration(duration_secs) -> str:
    """
    Formats duration in English.

    >>> format_sequence_duration(30)
    '30 seconds'
    >>> format_sequence_duration(90)
    '1 minute, 30 seconds'
    >>> format_sequence_duration(60 * 60 + 120)
    '1 hour, 2 minutes'
    """
    duration_mins = (duration_secs // 60) % 60
    duration_hrs = duration_secs // 60 // 60
    duration_secs = duration_secs % 60

    format_components = []
    if duration_hrs > 0:
        hrs = '{} {}'.format(duration_hrs, pluralize('hour', duration_hrs))
        format_components.append(hrs)

    if duration_mins > 0:
        mins = '{} {}'.format(
            duration_mins, pluralize('minute', duration_mins))
        format_components.append(mins)

    # Show seconds if there are no hours.
    if duration_hrs == 0:
        secs = '{} {}'.format(
            duration_secs, pluralize('second', duration_secs))
        format_components.append(secs)

    return ', '.join(format_components)


def esc(string):
    if string is None:
        return None
    return string.replace('&', '&amp;').replace(" target='_blank'", '')


def dot_join(*items):
    """
    Joins the given strings with a dot character. Filters out None values.
    """
    return '  •  '.join(map(str, filter(lambda x: x is not None, items)))


def get_cached_status_icon(cache_status: SongCacheStatus):
    cache_icon = {
        SongCacheStatus.NOT_CACHED: '',
        SongCacheStatus.CACHED: 'folder-download-symbolic',
        SongCacheStatus.PERMANENTLY_CACHED: 'view-pin-symbolic',
        SongCacheStatus.DOWNLOADING: 'emblem-synchronizing-symbolic',
    }
    return cache_icon[cache_status]


def _parse_diff_location(location):
    match = re.match(r'root\[(\d*)\](?:\[(\d*)\]|\.(.*))?', location)
    return tuple(g for g in match.groups() if g is not None)


def diff_song_store(store_to_edit, new_store):
    """
    Diffing song stores is nice, because we can easily make edits by modifying
    the underlying store.
    """
    old_store = [row[:] for row in store_to_edit]

    # Diff the lists to determine what needs to be changed.
    diff = DeepDiff(old_store, new_store)
    changed = diff.get('values_changed', {})
    added = diff.get('iterable_item_added', {})
    removed = diff.get('iterable_item_removed', {})

    for edit_location, diff in changed.items():
        idx, field = _parse_diff_location(edit_location)
        store_to_edit[int(idx)][int(field)] = diff['new_value']

    for add_location, value in added.items():
        store_to_edit.append(value)

    for remove_location, value in reversed(list(removed.items())):
        remove_at = int(_parse_diff_location(remove_location)[0])
        del store_to_edit[remove_at]


def diff_model_store(store_to_edit, new_store):
    """
    The diff here is that if there are any differences, then we refresh the
    entire list. This is because it is too hard to do editing.
    """
    old_store = store_to_edit[:]

    diff = DeepDiff(old_store, new_store)
    if diff == {}:
        return

    store_to_edit.remove_all()
    for model in new_store:
        store_to_edit.append(model)


def show_song_popover(
    song_ids,
    x: int,
    y: int,
    relative_to: Any,
    position: Gtk.PositionType = Gtk.PositionType.BOTTOM,
    on_download_state_change: Callable[[int], None] = lambda x: None,
    show_remove_from_playlist_button: bool = False,
    extra_menu_items: List[Tuple[Gtk.ModelButton, Any]] = [],
):
    def on_download_songs_click(button):
        CacheManager.batch_download_songs(
            song_ids,
            before_download=on_download_state_change,
            on_song_download_complete=on_download_state_change,
        )

    def on_remove_downloads_click(button):
        CacheManager.batch_delete_cached_songs(
            song_ids,
            on_song_delete=on_download_state_change,
        )

    def on_add_to_playlist_click(button, playlist):
        CacheManager.executor.submit(
            CacheManager.update_playlist,
            playlist_id=playlist.id,
            song_id_to_add=song_ids,
        )

    popover = Gtk.PopoverMenu()
    vbox = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)

    # Add all of the menu items to the popover.
    song_count = len(song_ids)

    # Determine if we should enable the download button.
    download_sensitive, remove_download_sensitive = False, False
    albums, artists, parents = set(), set(), set()
    for song_id in song_ids:
        details = CacheManager.get_song_details(song_id).result()
        status = CacheManager.get_cached_status(details)
        albums.add(details.albumId)
        artists.add(details.artistId)
        parents.add(details.parent)

        if download_sensitive or status == SongCacheStatus.NOT_CACHED:
            download_sensitive = True

        if (remove_download_sensitive
                or status in (SongCacheStatus.CACHED,
                              SongCacheStatus.PERMANENTLY_CACHED)):
            remove_download_sensitive = True

    go_to_album_button = Gtk.ModelButton(
        text='Go to album', action_name='app.go-to-album')
    if len(albums) == 1 and list(albums)[0] is not None:
        album_value = GLib.Variant('s', list(albums)[0])
        go_to_album_button.set_action_target_value(album_value)

    go_to_artist_button = Gtk.ModelButton(
        text='Go to artist', action_name='app.go-to-artist')
    if len(artists) == 1 and list(artists)[0] is not None:
        artist_value = GLib.Variant('s', list(artists)[0])
        go_to_artist_button.set_action_target_value(artist_value)

    browse_to_song = Gtk.ModelButton(
        text=f"Browse to {pluralize('song', song_count)}",
        action_name='app.browse-to',
    )
    if len(parents) == 1 and list(parents)[0] is not None:
        parent_value = GLib.Variant('s', list(parents)[0])
        browse_to_song.set_action_target_value(parent_value)

    menu_items = [
        Gtk.ModelButton(
            text='Play next',
            action_name='app.play-next',
            action_target=GLib.Variant('as', song_ids),
        ),
        Gtk.ModelButton(
            text='Add to queue',
            action_name='app.add-to-queue',
            action_target=GLib.Variant('as', song_ids),
        ),
        Gtk.Separator(orientation=Gtk.Orientation.HORIZONTAL),
        go_to_album_button,
        go_to_artist_button,
        browse_to_song,
        Gtk.Separator(orientation=Gtk.Orientation.HORIZONTAL),
        (
            Gtk.ModelButton(
                text=f"Download {pluralize('song', song_count)}",
                sensitive=download_sensitive,
            ),
            on_download_songs_click,
        ),
        (
            Gtk.ModelButton(
                text=f"Remove {pluralize('download', song_count)}",
                sensitive=remove_download_sensitive,
            ),
            on_remove_downloads_click,
        ),
        Gtk.Separator(orientation=Gtk.Orientation.HORIZONTAL),
        Gtk.ModelButton(
            text=f"Add {pluralize('song', song_count)} to playlist",
            menu_name='add-to-playlist',
        ),
        *extra_menu_items,
    ]

    for item in menu_items:
        if type(item) == tuple:
            el, fn = item
            el.connect('clicked', fn)
            el.get_style_context().add_class('menu-button')
            vbox.pack_start(item[0], False, True, 0)
        else:
            item.get_style_context().add_class('menu-button')
            vbox.pack_start(item, False, True, 0)

    popover.add(vbox)

    # Create the "Add song(s) to playlist" sub-menu.
    playlists_vbox = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)

    # Back button
    playlists_vbox.add(
        Gtk.ModelButton(
            inverted=True,
            centered=True,
            menu_name='main',
        ))

    # The playlist buttons
    for playlist in CacheManager.get_playlists().result():
        button = Gtk.ModelButton(text=playlist.name)
        button.get_style_context().add_class('menu-button')
        button.connect('clicked', on_add_to_playlist_click, playlist)
        playlists_vbox.pack_start(button, False, True, 0)

    popover.add(playlists_vbox)
    popover.child_set_property(playlists_vbox, 'submenu', 'add-to-playlist')

    # Positioning of the popover.
    rect = Gdk.Rectangle()
    rect.x, rect.y, rect.width, rect.height = x, y, 1, 1
    popover.set_pointing_to(rect)
    popover.set_position(position)
    popover.set_relative_to(relative_to)

    popover.popup()
    popover.show_all()


def async_callback(
    future_fn,
    before_download=None,
    on_failure=None,
):
    """
    Defines the ``async_callback`` decorator.

    When a function is annotated with this decorator, the function becomes the
    done callback for the given future-generating lambda function. The
    annotated function will be called with the result of the future generated
    by said lambda function.

    :param future_fn: a function which generates a
        ``concurrent.futures.Future``.
    """
    def decorator(callback_fn):
        @functools.wraps(callback_fn)
        def wrapper(self, *args, state=None, **kwargs):
            if before_download:
                on_before_download = (
                    lambda: GLib.idle_add(before_download, self))
            else:
                on_before_download = (lambda: None)

            def future_callback(f):
                try:
                    result = f.result()
                except Exception as e:
                    if on_failure:
                        on_failure(self, e)
                    return

                return GLib.idle_add(callback_fn, self, result, state)

            future: Future = future_fn(
                *args,
                before_download=on_before_download,
                **kwargs,
            )
            future.add_done_callback(future_callback)

        return wrapper

    return decorator
